#include <stdio.h> 
#include <math.h> 
#include <signal.h>
#include "backtrack.h"
#include "part.h"
#include "data_sets.h"

void line_randomize();
void randomize_vertices();
int find_max_degree();
int compute_initial_cost(int [][MAX_VERTICES], int *);

void perform_maximize(int, int, int*);
void post_processing(int, int, int);
void test_nongreedy_placement(int, int *);
void test_greedy_placement(int, int *);
void update_push_history(int node, int side);
void free_push_history();
void print_push_history(int *vertices);
int check_cost(int *, int);
int variance_stats_ok(int *, int);
int compute_cost(int *);
int stats_say_ng_step(int, int);
void init_push_history();
void get_config();
void find_augmenting_paths();
void fix_walk_subset(int, int);
void fix_kl_subset(int low, int high);
void kl_init_loop(int);
void kl_main_loop();


#define NONGREEDY_PROB         	0.0005	/** For W-non with certain prob.    **/

extern long lrand48();

extern int num ;              		/** number of vertices 		    **/
extern int mode;

extern int hours;
extern double last_time;
extern int deg[];     			/** number of edges for each vertex **/
extern int Exhaustive_Search;		/** bk: exhaustive or not?	    **/
extern int NG_Max;			/** bk: exh, but only NG_Max ng.    **/
extern int Augment_With_Local;		/** run local opt after each walk   **/
extern int Augment_With_KL;		/** run KL once   after each walk   **/
extern int Augment_With_Paths;	

extern int node_cost[2][MAX_VERTICES]; 	/** cost of vertex in each partition**/
extern int vertices[MAX_VERTICES] ;
extern int num_vertices[2];		/** for kl			   **/
extern int hold_num_vertices[2];	/** for kl			   **/
extern int hold_vertices[MAX_VERTICES];
extern int subgraph_deg[MAX_VERTS];
extern int currentcost; 
extern int currentminsize; 
extern double time_;
extern PG_DIST (*ng_distrib)[DISTRIB_SIZE];/** lookup tab used by P_Heuristic**/

extern FILE *outfile;
extern struct adj_node *adj_list[MAX_VERTICES] ;

int No_BK_Reorder;			/** flag - use dumb selection?	    **/
int Near_Greedy;			/** flag - use near greedy somehow? **/
int N_Interval;				/** num v in 1% of NG_UP_Results    **/
int P_Interval;				/** num v in 1% 		    **/
int P_Heuristic;			/** flag - use P heuristic? 	    **/
int P_Prob;				/** which row of ng_distrib to use  **/

int stack_ptr;
int last_part_diff;
int max_part_diff;
int ng_cost[2][MAX_VERTS]; 		/** node cost (in terms of ng vert)**/
int hold_ncost[2][MAX_VERTICES];	/** for init'ing kl		   **/

int vert_processed[MAX_VERTICES]; 	/** unprocessed vertices           **/
int vert_loc[MAX_VERTICES];       	/** vertex location                **/
int num_verts[2]; 
int hold_num_verts[2];
int hold_loc[MAX_VERTICES];         	/** temp. storage for best solutions **/
int total_cands;		    	/** backtracking candidates	     **/
int cand_calls;		    	    	/** number of times 'find_cand..()'  **/
int big_cands;			    	/** times > 2 cands in 1 call	     **/

int tot_cost;				/** the current cumulative #cuts     **/
int ng_count;				/** the number of nongreedy steps    **/
					/** taken at the current point in    **/
					/** the computation.		     **/

int dist_trace[SIZE_INTERVAL_ARRAY];	/** how many times was each distr.   **/
					/** used during a PG2 run?           **/
int hold_pg_ng_dist[MAX_VERTS];	/** for best walk, recrd n/ng all stp.*/
int pg_ng_dist[MAX_VERTS];		/** for each walk, recrd n/ng all stp.*/
int PG_Distribution[MAX_VERTS];	/** sum of best walks of all graphs   */
/******************************************************************************/
/** only useful in experiments with 500 vertices ******************************/
/******************************************************************************/
int ng_before_100;			/** #ng before 100th placement	     **/
int ng_before_250;			/** #ng before 250th placement	     **/
int ng_before_400;			/** #ng before 400th placement	     **/

int Hold_NG_Count;			/** best ng_count of k walks	     **/
int Hold_NG_Before_100;		
int Hold_NG_Before_250;	
int Hold_NG_Before_400;

int num_processed;			/* num vertices placed in a part. */
/*int abs_diff[MAX_VERTICES];*/		/* partitioning cost	          */
int Stack_size;		/* for debugging only */
int bk_cost ;
int bk_minsize ;

double total_part_diff;
double ave_cands;			/** total cands/ # selections	     **/
double Hold_Ave_Cands;			/** best of k walks		     **/
double leaf_count;			/* leaves in backtracking tree       **/

short bits[(int) MAX_NG_STRING];	/** for use with -PC (configs) option**/

STK_ELEMPTR	stack;


void push(int type, int value, int side, int dist, int greediness)
{
	extern int stack_ptr;
	extern int Print_Cand_History;

	if (stack_ptr == Stack_size) {
/*#ifndef CC_COMPILER
 		stack = (STK_ELEMPTR) realloc((malloc_t)stack,(2 * Stack_size) *
 						     sizeof(STK_ELEM));
#else*/
		stack = (STK_ELEMPTR) realloc((void *)stack,(2 * Stack_size) *
						     sizeof(STK_ELEM));
/*#endif*/
		Stack_size *= 2;
		printf("STACK DOUBLED\n");
	}

	stack[stack_ptr].type = type;
	stack[stack_ptr].value = value;
	stack[stack_ptr].side = side;
	stack[stack_ptr].dist = dist;
	stack[stack_ptr].greediness = greediness;
	/*vert_loc[value] = side;*/     /** why? **/
	stack_ptr++;
	if (type == CAND && Print_Cand_History)
		update_push_history(value, side);
}

int pop()
{
	extern int ng_count;

	if (stack_ptr == 0)
		return(BOS);			/* BOS is a large neg. */
	if (stack[stack_ptr].greediness == NONGREEDY)	/** ?? 5/4/92 **/
		ng_count--;
	stack_ptr--;
	return(stack_ptr);
}

int pop_and_clear()
{
	int i;

	if (stack_ptr == 0) {
		printf("Empty Stack error (num_processed: %d)\n",num_processed);
		;
	}
	/*i = (int) (random() % stack_ptr);*/ 	/* 4/27/93 */
	i = (int) (lrand48() % stack_ptr);

	if (stack[i].greediness == NONGREEDY) {
		ng_count++;
		pg_ng_dist[num_processed] = 1;
		if (num_processed < 100) ng_before_100++;
		if (num_processed < 250) ng_before_250++;
		if (num_processed < 400) ng_before_400++;
	}
	stack_ptr = 0;
	/*printf("actual choice: %d\n", stack[i].value);*/
	return(i);
}

int find_part_diff()
{
	static int last_loc[MAX_VERTS];
	int i;

	if ((int) leaf_count == 1)
		for (i=0; i<num; i++)
			last_loc[i] = vert_loc[i];
	last_part_diff = num;
	for (i=0; i<num; i++)
		last_part_diff -= (last_loc[i] == vert_loc[i]);
	if (last_part_diff > (num-last_part_diff))
		last_part_diff = num-last_part_diff;
	if (last_part_diff > max_part_diff)
		max_part_diff = last_part_diff;
	for (i=0; i<num; i++)
		last_loc[i] = vert_loc[i];
	return last_part_diff;
}

void save_state_if_improvement(int local)
{
	extern LISTPTR nodes;
	extern int Heuristic_Only;
	int i;
	double current_quotient=0.0, best_quotient=0.0;

        if (mode == MIN_QUOTIENT) {
                current_quotient = (double)currentcost/(double)currentminsize;
                best_quotient   = (double)bk_cost  /(double)bk_minsize;
        }
	if ( ((mode == MODE_MAX) && (currentcost > bk_cost)) ||
	     ((mode == MODE_MIN) && (currentcost < bk_cost)) ||
	     ((mode == MIN_QUOTIENT) && (current_quotient < best_quotient)) ) {
		bk_cost = currentcost;
		bk_minsize = currentminsize;
		if (!local) {
			for (i=0; i<num; i++) {
				hold_loc[i] = vert_loc[i];
				hold_vertices[i] = vertices[i];
				hold_ncost[SET_A][i] = node_cost[SET_A][i];
				hold_ncost[SET_B][i] = node_cost[SET_B][i];
			}
			hold_num_verts[SET_A] = num_verts[SET_A];
			hold_num_verts[SET_B] = num_verts[SET_B];
		} else {
			for (i=0; i<num; i++) {
				hold_loc[i] = nodes[i].loc;
				hold_vertices[i] = vertices[i];
			}
			hold_num_verts[SET_A] = num_vertices[SET_A];
			hold_num_verts[SET_B] = num_vertices[SET_B];
			if (!Heuristic_Only && !check_cost(hold_loc, bk_cost))
				printf("******program error 6\n");
		}
		Hold_Ave_Cands = ave_cands;
		Hold_NG_Count = ng_count;
		/*for (i=0; i<num; i++) 
			hold_pg_ng_dist[i] = pg_ng_dist[i];*/
		Hold_NG_Before_100 = ng_before_100;
		Hold_NG_Before_250 = ng_before_250;
		Hold_NG_Before_400 = ng_before_400;
	}
}

void compute_costs3(int *locs, int *processed)
{
        struct adj_node *point;
        int i;

        for(i=0;i<num;i++) {
                node_cost[SET_A][i] = node_cost[SET_B][i] = 0;
                for (point = adj_list[i]; point; point = point->next)
	   	    if (processed[point->node_no]) {
                        if(vert_loc[point->node_no] == SET_A)
                                node_cost[SET_B][i]++;
                        else 
                                node_cost[SET_A][i]++;
		    }
        }
}

void check_nongreedy(int v, int *side)
{
	extern int num;
	extern int total_greedy_cost[MAX_VERTS];   /* see greedy_history.c */
	extern int total_nongreedy_cost[MAX_VERTS];
	extern int greedy_changes[MAX_VERTS];	
	extern int nongreedy_changes[MAX_VERTS];
	extern int class_[MAX_VERTS];	
	extern int greedy_class[MAX_VERTS];	
	extern int nongreedy_class[MAX_VERTS];
	extern LISTPTR nodes;
	int i, node;
	int processed[MAX_VERTS];
	struct adj_node *point;

	for (i=0; i<num; i++) {
		class_ [i] = greedy_class[i] = nongreedy_class[i] = 0;
		greedy_changes[i] = nongreedy_changes[i] = 0;
		total_greedy_cost[i] = total_nongreedy_cost[i] = 0;
		processed[i] = nodes[i].lock;
	}
	compute_costs3(vert_loc, processed);
	if ((*side == SET_A && node_cost[SET_B][v] > node_cost[SET_A][v])||
	    (*side == SET_B && node_cost[SET_A][v] >=node_cost[SET_B][v]))
		printf("ERROR in check_nongreedy\n");

	for (point = adj_list[v]; point; point = point->next) {
		node = point->node_no;
		if (!processed[node])
		   class_[node]=node_cost[SET_A][node]-node_cost[SET_B][node];
	}
	test_greedy_placement(v, processed);
	test_nongreedy_placement(v, processed);

        if ((mode==MODE_MAX && total_greedy_cost[v]>=total_nongreedy_cost[v])||
            (mode==MODE_MIN && total_greedy_cost[v]<=total_nongreedy_cost[v]))
		return;
	else
		*side = !(*side);
	return;
}

int refine(int value, int side, int dist, int operation, int greediness)
{
	extern int PG_VG_Flag;
	extern int Change_Index;
	extern int currentminsize;
	extern int ng_count;
	extern int tot_cost;
	extern LISTPTR nodes;
	int vnum;			/* was static for a time */
	static struct adj_node *point;
	static LISTPTR temp;
	int current_var_stats[NUM_DATA_SETS];
	/*int free_placements;
	static int total_free_placements;*/

	if (operation == PLACE) {
		/****************************************************/
		/** EXTREMELY SLOW!  MAKE SURE TO REMOVE THIS LINE **/
		/****************************************************/
		/*check_nongreedy(value, &side);*/
		/****************************************************/
		/***********************************************************/
		/*** test bad placement of a single vertex *****************/
		/***********************************************************/
		if (num_processed+1 == Change_Index)
			side = !side;
		/***********************************************************/
		for (point = adj_list[value]; point; point = point->next) {
			if (nodes[point->node_no].lock != LOCKED) {
				vnum = point->node_no;
				temp = &nodes[vnum];
				if (side == SET_A) {
					move_to_bucket(0, temp->cost+1, temp);
					node_cost[SET_B][vnum]++;
					if (greediness == NONGREEDY)
					    ng_cost[SET_B][vnum]++;
				} else {
					move_to_bucket(0, temp->cost-1, temp);
					node_cost[SET_A][vnum]++;
					if (greediness == NONGREEDY)
					    ng_cost[SET_A][vnum]++;
				}
				/*subgraph_deg[vnum]--;
				free_placements = 0;
				if (ABS(temp->cost) > subgraph_deg[vnum]) {
					total_free_placements++;
					free_placements++;
					test_and_push(vnum);
					printf("###%d\n",vnum);
				}*/
			}
		}
		/*printf("ddiff(%d) = %d\n", num_processed, 
				ABS(nodes[value].cost)-subgraph_deg[value]);*/
		dist_trace[dist]++;		/* for PG2 only */
		nodes[value].lock = LOCKED;
		vert_loc[value]  = side;
		delete_from_bucket(&nodes[value]);
		num_verts[side]++;
		currentminsize = MIN(num_verts[0], num_verts[1]);
		tot_cost += node_cost[side][value];
		vertices[num_processed] = value;
		if (PG_VG_Flag) {
			current_var_stats[NG_COUNT] = ng_count;
			current_var_stats[TOT_COST] = tot_cost;
			current_var_stats[SIZE_A] = num_verts[SET_A];
			current_var_stats[SIZE_B] = num_verts[SET_B];
			if (variance_stats_ok(current_var_stats,num_processed)
								== FAIL)
				return(FAIL);
		}
		/*while (free_placements-- > 0)
			select(RANDOM_WALK);*/
		if (++num_processed == num) {
		   /*printf("total_free_plcmnt: %d\n",total_free_placements);*/
			leaf_count = leaf_count + 1;
			if (tot_cost != (currentcost = compute_cost(vert_loc)))
				printf("ERROR: refine()\n"), exit(4);
			if (!Augment_With_Local && !Augment_With_KL && 
			    !Augment_With_Paths)
				save_state_if_improvement(!AUGMENT);
		}
	} else if (operation == UNPLACE) {
		for (point = adj_list[value]; point; point = point->next) {
			if (nodes[point->node_no].lock != LOCKED) {
				vnum = point->node_no;
				temp = &nodes[vnum];
				if (side == SET_A) {
					move_to_bucket(0, temp->cost-1, temp);
					node_cost[SET_B][vnum]--;
					if (greediness == NONGREEDY)
					    ng_cost[SET_B][vnum]--;
				} else {
					move_to_bucket(0, temp->cost+1, temp);
					node_cost[SET_A][vnum]--;
					if (greediness == NONGREEDY)
					    ng_cost[SET_A][vnum]--;
				}
			}
		}		
		dist_trace[dist]--;/* for PG2 only */
		nodes[value].lock = NOT_LOCKED;
		insert_to_bucket(0, nodes[value].cost, &nodes[value]);
		num_verts[side]--;
		num_processed--;
	}
	return(PASS);
}

void print_nodes()
{
	extern LISTPTR nodes;
	int i;

	printf("\n*****************************************************\n");
	for (i=0; i<num; i++)
		printf("(%d, deg: %d, loc: %d, lock: %d)\n",nodes[i].node_no, 
				deg[nodes[i].node_no], nodes[i].loc,
				nodes[i].lock);
	printf("\n*****************************************************\n");
}


int select(int stack_protocol)		/** param: current node ?? **/
{
	int i;

	if (stack_protocol == BACKTRACK) {
		if ((i=pop()) == BOS)	return(BOS);
	} else if (stack_protocol == RANDOM_WALK)
		i = pop_and_clear();
	else    printf("stack protocol error\n"),exit(5);

	switch (stack[i].type) {
		case CAND:
				if (refine(stack[i].value, stack[i].side,
					stack[i].dist, 
					PLACE, stack[i].greediness) == FAIL)
						return(FAIL);
				break;
		case UP:	while (stack[i].type == UP) {
				   refine(stack[i].value,stack[i].side,
					  stack[i].dist,
					  UNPLACE, stack[i].greediness);
				   if ((i=pop()) == BOS)	return(BOS);
				}
				refine(stack[i].value, stack[i].side, 
					stack[i].dist,
					PLACE, stack[i].greediness);
				break;
		default:	printf("error in select()\n"),exit(1);
	}
	if (stack_protocol == BACKTRACK) 
		push(UP, stack[i].value, vert_loc[stack[i].value], 0,
					 stack[i].greediness);
	return(stack[i].value);
}

/******************************
stack_candidate(int type, int node, int side)
{
	double e;

	if ((e=rrandom()) < NONGREEDY_PROB) {
		printf("rrandom(): %5.3f\n", e);
		push(type, node, !side);
	} else
		push(type, node, side);
}
**********************************/


void stack_candidate(int type, int node, int side)
{
	extern int Use_Configurations;
	extern int Use_Fixed_Prob;
	extern int Use_One_Distribution;
	extern int Abs_Diff_Intervals;
	extern int Up_NG_Intervals;
	extern double find_next_ng_prob(int, int);
	/*extern double find_next_ng_prob(int, int);*/
	extern int PG_VG_Flag2;
	extern int VG_Flag;
	/*extern int Leaves;*/

	int dist_used;
	double ng_prob;
	double e;

	/*if ( (num / (Leaves+1)) == num_processed)
		printf("hi mom\n");*/
	if (Exhaustive_Search) {
		if (ng_count < NG_Max) {
			push(type, node, !side, 0, NONGREEDY);
			push(type, node, side, 0, GREEDY);
			ng_count++;
		} else
			push(type, node, side, 0, GREEDY);
	} else if (P_Heuristic) {
		if (Use_Configurations && (num_processed <= MAX_NG_STRING)) {
			if (bits[num_processed] == 0)
				push(type, node, !side, dist_used, GREEDY);
			else
				push(type, node, !side, dist_used, NONGREEDY);
			return;
		}

		if (Use_Fixed_Prob)
			ng_prob = fixed_prob();
		else if (Use_One_Distribution)
			ng_prob = find_prob(node, &dist_used);
		else if (Abs_Diff_Intervals)
			ng_prob = find_prob_abs(node, &dist_used);
		else if (Up_NG_Intervals)
			ng_prob = find_prob_k(node, &dist_used);
		else
			ng_prob = find_prob_k_over_total(node, &dist_used);

		if ((e=rrandom()) < ng_prob) {
			push(type, node, !side, dist_used, NONGREEDY);
		} else
			push(type, node, side, dist_used, GREEDY);
	} else if (VG_Flag) {  /* deterministic alg */
		if (stats_say_ng_step(ng_count, num_processed))
			push(type, node, !side, 0, NONGREEDY);
		else
			push(type, node, side, 0, GREEDY);
	} else if (PG_VG_Flag2) {  /* guided, but not restarting pg-vg alg */
		if ((e=rrandom())<(ng_prob=find_next_ng_prob(ng_count, 
							     num_processed))){
			push(type, node, !side, 0, NONGREEDY);
		} else
			push(type, node, side, 0, GREEDY);
	} else
		push(type, node, side, 0, GREEDY);
}

void test_and_push(int vert_num)
{
	extern LISTPTR nodes;

	int diff;

	if (mode == MODE_MAX) {
		if (nodes[vert_num].cost <= 0) 
			push(CAND, vert_num, SET_A, 0, GREEDY);
		else
			push(CAND, vert_num, SET_B, 0, GREEDY);
	} else {
		diff = num_verts[SET_A] - num_verts[SET_B];
		if ( ((nodes[vert_num].cost<=0) && (-diff<=UNBALANCE))
						|| (diff >= UNBALANCE) )
			push(CAND, vert_num, SET_B, 0, GREEDY);
		else
			push(CAND, vert_num, SET_A, 0, GREEDY);
	}
}

void test_and_stack(int vert_num)
{
	extern LISTPTR nodes;

	int new_cost0, new_cost1;
	int new_size0, new_size1;
	double new_quot0, new_quot1;
	int diff;

	if (mode == MODE_MAX) {
		if (nodes[vert_num].cost <= 0) 
			stack_candidate(CAND, vert_num, SET_A);
		else
			stack_candidate(CAND, vert_num, SET_B);
	} else if (mode == MODE_MIN) {
		diff = num_verts[SET_A] - num_verts[SET_B];
		if ( ((nodes[vert_num].cost<=0) && (-diff<=UNBALANCE))
						|| (diff >= UNBALANCE) )
			stack_candidate(CAND, vert_num, SET_B);
		else
			stack_candidate(CAND, vert_num, SET_A);
	} else if (mode == MIN_QUOTIENT) {
		new_cost0 = tot_cost + node_cost[0][vert_num];
		new_cost1 = tot_cost + node_cost[1][vert_num];
		new_size0 = MIN(num_verts[0]+1, num_verts[1]);
		new_size1 = MIN(num_verts[1]+1, num_verts[0]);
                if (new_size0 > 0)
                        new_quot0 = ((double) new_cost0 / (double) new_size0);
                else
                        new_quot0 = MAX_INT;
                if (new_size1 > 0) 
                        new_quot1 = ((double) new_cost1 / (double) new_size1);
                else
                        new_quot1 = MAX_INT;

                if (new_quot0 < new_quot1)
			stack_candidate(CAND, vert_num, SET_A);
		else if (new_quot0 > new_quot1)
			stack_candidate(CAND, vert_num, SET_B);
		else {
			if (new_size0 <= new_size1)
				stack_candidate(CAND, vert_num, SET_B);
			else
				stack_candidate(CAND, vert_num, SET_A);
		}
	}
}

int find_candidates(int mode, int placement_number)
{
	extern 	int 	Many_Cands;
	extern 	LISTPTR *buckets[2];
	extern	int	Current_Max_Bucket[2], Current_Min_Bucket[2];
	extern	int	Num_Pos_Buckets;
	int	i,
		tempptr,
		cand_found = FALSE,
		new_cands = 0,
		max_degree = 0,
		min_degree = MAX_INT,
		temp[MAX_VERTICES],
		min_cost,
		max_cost,
		mode_flag = (mode == MODE_MAX);
	LISTPTR bucket, node;

	if (No_BK_Reorder) {
		test_and_stack(placement_number);
		return(1);
	}

	tempptr = 0;

	/********************************************************************/
	/** only buckets[0] is used in this case since unplaced vertices   **/
	/** don't have a partition by definition.  buckets[1] comes into   **/
	/** use with KL and with local optimization.                       **/
	/********************************************************************/
        if (buckets[0][Current_Max_Bucket[0]]->next == NULL) {
                for (i=Current_Max_Bucket[0]; i >= Current_Min_Bucket[0]; i--)
                        if (buckets[0][i]->next != NULL)
                                break;
                Current_Max_Bucket[0] = i;
        }

        if (buckets[0][Current_Min_Bucket[0]]->next == NULL) {
                for (i=Current_Min_Bucket[0]; i <= Current_Max_Bucket[0]; i++)
                        if (buckets[0][i]->next != NULL)
                                break;
                Current_Min_Bucket[0] = i;
        }
	min_cost = Current_Min_Bucket[0] - Num_Pos_Buckets;
	max_cost = Current_Max_Bucket[0] - Num_Pos_Buckets;

	if (((mode == MODE_MAX) && (ABS(min_cost) >= max_cost)) ||
	    ((mode != MODE_MAX) && (num_verts[SET_A] >= num_verts[SET_B]))) {
		bucket = buckets[0][Current_Min_Bucket[0]];
		for(node = bucket->next; node /*&& (tempptr < MAX_CANDS)*/; 
						node = node->next) {
			temp[tempptr++] = node->node_no;
			if (mode_flag && (deg[node->node_no] > max_degree))
				max_degree = deg[node->node_no];
			if (!mode_flag && (deg[node->node_no] < min_degree))
				min_degree = deg[node->node_no];
		}
	} 
	if ((((mode == MODE_MAX) && (ABS(min_cost) <= max_cost)) &&
	    (Current_Min_Bucket[0] != Current_Max_Bucket[0])) ||
	     ((mode != MODE_MAX) && (num_verts[SET_A] < num_verts[SET_B]))) {
		bucket = buckets[0][Current_Max_Bucket[0]];
		for(node = bucket->next; node /*&& (tempptr < MAX_CANDS)*/; 
						node = node->next) {
			temp[tempptr++] = node->node_no;
			if (mode_flag && (deg[node->node_no] > max_degree))
				max_degree = deg[node->node_no];
			if (!mode_flag && (deg[node->node_no] < min_degree))
				min_degree = deg[node->node_no];
		}
	} 

	for (i=0; i < tempptr; i++) {
		if (Many_Cands || 
		    ((mode_flag && (deg[temp[i]] == max_degree)) ||
		     (!mode_flag && (deg[temp[i]] == min_degree)))) {
		   cand_found = TRUE;
		   test_and_stack(temp[i]);
		   new_cands++;
		   total_cands++;
		}
		if ((mode_flag && (max_degree == 0)) || 
		    (!mode_flag && (min_degree == MAX_INT))||Exhaustive_Search) 
			break;
	}
	if (cand_found) cand_calls++;
	if (new_cands >= 2) big_cands++;	
	/*printf("%d (%d,%d,%d) (%d) ", new_cands, 
			min_cost, max_cost,min_degree, tempptr);*/
	return(new_cands);
}

void init_walk(int reset_locs)
{
	extern int Use_Configurations;
	extern int KL_PG_Flag;
	/* extern int pg_ng_dist[]; */
	extern int ng_count;
	extern int tot_cost;
	extern int ng_before_100, ng_before_250, ng_before_400;
	extern int Print_Cand_History;
	int i, maxdeg = 0, mindeg = MAX_INT, first = 0;
	int mode_flag = (mode == MODE_MAX);

	currentminsize = 1;
	ng_count = 0;
	tot_cost = 0;
	ng_before_100 = 0;
	ng_before_250 = 0;
	ng_before_400 = 0;
	num_processed = 0;
	cand_calls = total_cands = big_cands = 0;
	num_verts[SET_A] = num_verts[SET_B] = 0;
	init_buckets(reset_locs);
	if (Print_Cand_History)
		init_push_history();
	for (i=0; i<SIZE_INTERVAL_ARRAY; i++) 
		dist_trace[i] = 0;		/* PG2 only */
	for (i=0; i<num; i++) {
		/*pg_ng_dist[i] = 0;*/
		/*subgraph_deg[i] = deg[i];*/
		vert_processed[i] = FALSE;
		node_cost[0][i] = node_cost[1][i] = 0;
		/*ng_cost[0][i] = ng_cost[1][i] = 0;*/
		/*abs_diff[i] = 0;*/
		if (mode_flag && (deg[i] > maxdeg)) {	
			maxdeg = deg[i];
			first = i;
		}
		if (!mode_flag && (deg[i] < mindeg)) {
			mindeg = deg[i];
			first = i;
		}
	}
	if (Use_Configurations)
		get_config();
	if (!KL_PG_Flag)
		push(CAND, first, SET_A, 0, GREEDY);
}

void init_backtrack(int create)
/**** called only once per graph ****/
{
	extern int num;
	extern int mode;
	extern int Total_Flip_Cost;
	extern int Total_Path_Length;
	extern int Num_Augmenting_Paths;
	extern int Total_Flip_Cost;
	extern int numedges;
	extern int bk_cost;
	extern int Paths;
	extern double tottime;
	int maxdeg;

	if (create) {
        	create_graph();
		maxdeg = find_max_degree();
        	/*alloc_buckets(2*num-1);*/
        	alloc_buckets(2*maxdeg+1);
	}
        process_reorder();
        if (num <= 20)  print_adj_list();
        tottime = 0.0;
        if(mode == MODE_MAX)
                bk_cost = VERY_LOW;
        else
                bk_cost = VERY_HIGH;
	bk_minsize = 1;
        fprintf(outfile,"OUTPUT:\n");
        fprintf(outfile,"   edges:%6d\n",numedges);
        if(mode == MODE_MAX)
                fprintf(outfile,"Maximize number of edges cut\n");
        else
                fprintf(outfile,"Minimize number of edges cut\n");

	max_part_diff = 0; 
	total_part_diff = 0;
	hours = 0;
	last_time = 0.0;
	Total_Path_Length = 0;
	Num_Augmenting_Paths = 0;
	Total_Flip_Cost = 0;

	Stack_size = INITSTACKSIZE;
	stack = (STK_ELEMPTR) calloc(INITSTACKSIZE, sizeof(STK_ELEM));

	/*signal(SIGBUS, handler);*/
	stack_ptr = 0;
	leaf_count = 0.0;
	if (Paths == BACKTRACK)
		init_walk(RESET_LOCS);
}

void fix_walk_subset(int low, int high)
{
	int i, v;

	for (i=low; i<high; i++) {
		v = vertices[i];
		refine(v, vert_loc[v], 0, PLACE, GREEDY);
	}
}

void fix_kl_subset(int low, int high)
{
	extern int ng_count;
	extern int hold_post_tots[];
	extern int hold_vertex_loc[];
	extern int hold_vertices[];
	int i, v;

	for (i=low; i<high; i++) {
		v = hold_vertices[i];
		if (hold_post_tots[i] < 0) {
			ng_count++;
			refine(v, hold_vertex_loc[v], 0, PLACE, NONGREEDY);
		} else
			refine(v, hold_vertex_loc[v], 0, PLACE, GREEDY);
	}
}

int one_walk(int low, int high, double *leaf_est)
{
	extern int num_processed;
	extern int W_Starts;
	extern int vert_loc[];
	extern int Starting_Locs[];
	int i=low, num_cands;

	/*for (i=low; i<high; i++) {*/
	while (num_processed < num) {
		if (select(RANDOM_WALK) == FAIL)
			return(FAIL);
		if ((num_cands = find_candidates(mode, i)) > 1)
			*leaf_est *= num_cands;
		i++;
	}
	if (W_Starts)
		for (i=0; i<num; i++) 
			Starting_Locs[i] = vert_loc[i];

	return (PASS);
}

void print_pass(int cost, int minsize, double ave_cands, double leaf_estimate, 
	   int big_cands, int num_verts_a, int num_verts_b)
{
	/* extern int Heuristic_Only; */
	extern FILE *outfile;
	char   solution[21] = "                    ";

	if (MIN_QUOTIENT)
		sprintf(solution,"%10.6f (%d)",(double) cost/(double) minsize,
                                  		cost);
	else
		sprintf(solution,"%-20d", cost);

	/*if (!Heuristic_Only)
		total_part_diff += (double) find_part_diff();*/

	fprintf(outfile,"%-20s%-10.3f%-9.0e%-9d%-9d%-6d%-4d\n",
		solution, 
		ave_cands,
		leaf_estimate,
		big_cands,
		last_part_diff,
       		num_verts_a,num_verts_b);
}

int explore_one_path(int mode, int print, double *leaf_estimate, int threshold)
/***************************************************************************/
/************* checks may be removed since an ultimate check is done *******/
/************* when the final result comes out.			     *******/
/***************************************************************************/
{
	extern int Heuristic_Only;
	extern int Print_Cand_History;
	extern int Paths;
	extern int Up_Down_Reps;
	int count = Up_Down_Reps; 

	if (one_walk(threshold, num, leaf_estimate) == FAIL)
		return(FAIL);

	if (print) {
		fprintf(outfile,"*");
		print_pass(currentcost, currentminsize, ave_cands, 
			   *leaf_estimate, big_cands,
			   num_verts[SET_A], num_verts[SET_B]);
	}
	while (count > 0) {
		int lower = 3 * num/4;
		int upper = num/4;

		init_walk(DONT_RESET_LOCS);
		if (count % 2  == 0) {
			fix_walk_subset(0, lower);
			find_candidates(mode, num_processed);
			if (one_walk(lower, num, leaf_estimate) == FAIL)
				return(FAIL);
		} else {
			fix_walk_subset(upper, num);
			find_candidates(mode, num_processed);
			if (one_walk(0, upper, leaf_estimate) == FAIL)
				return(FAIL);
		}
		count--;
	    	if (print)
		 /*print_pass(bk_cost, bk_minsize, Hold_Ave_Cands, 
			      *leaf_estimate, big_cands,
			      hold_num_verts[SET_A], 
			      hold_num_verts[SET_B]);*/
		 print_pass(currentcost, currentminsize, ave_cands, 
			   *leaf_estimate, big_cands,
			   num_verts[SET_A], 
			   num_verts[SET_B]);
	}
	if (print)
		fprintf(outfile, "\n********\n");

	ave_cands = (double) total_cands / (double) cand_calls;

	if (Near_Greedy)
	   fprintf(outfile,"(nongreedy: %-6d)\n", ng_count);

	if (!Heuristic_Only && !check_cost(vert_loc, currentcost))
			printf("program error1**********************\n");
	if (Augment_With_Local || Augment_With_Paths)
		init_local_opt(vert_loc, num_verts);
	if (Augment_With_Local) {
		local_opt();
		if (!Heuristic_Only && !check_cost(NULL, currentcost))
			printf("program error2**********************\n");
	}  else if (Augment_With_KL) {
		kl_augment(vert_loc);
		if (!Heuristic_Only && !check_cost(NULL, currentcost))
			printf("program error4**********************\n");
	}
	if (Augment_With_Paths) {
		find_augmenting_paths();
		if (!Heuristic_Only && !check_cost(NULL, currentcost))
			printf("program error2896**********************\n");
	}
	if (Paths != BACKTRACK && Print_Cand_History) {
		print_push_history(vertices);
		free_push_history();
	}
	/*****************************************************************/
	/*print_locs(vert_loc);*/
	/*****************************************************************/

	return(PASS);
}


double greedy_tree_walk(int mode, int paths, int print, int threshold)
{
	extern int KL_PG_Flag;
	extern int Paths;
	extern int Heuristic_Only;
	extern double tottime;
	extern double cpu_seconds;
	int i=0, j, more_work = TRUE; 
	double leaf_estimate, total_leaf_estimate = 0.0;
	static char hyphens[] = "--------------------------------------------------------------------------------";

	/*for (i=0;i<num;i++) 
		hold_pg_ng_dist[i] = 0;*/
	if (print) {
        	fprintf(outfile,"%-66s\n", hyphens);
        	fprintf(outfile,"%-6s%-10s%-9s%-9s%-10s%-6s%-4s\n", "CUTS",  "AVE.CANDS", "LEAVES",
			"TIMES>=2", "PART.DIFF.", "SET_A", "SET_B");
        	fprintf(outfile,"%-66s\n", hyphens);
	}
	/**********************************************************/
	/********************* cpu-time option here  **************/
	/**********************************************************/
	i=0;
	while (more_work) {
		leaf_estimate = 1.0;
		time_ = log_time();
		init_walk(RESET_LOCS);
		if (KL_PG_Flag) {
			fix_kl_subset(0, threshold);
			find_candidates(mode, num_processed);
		}
		if (explore_one_path(mode, print, &leaf_estimate, threshold) 
							== FAIL)
			continue;
		time_ = log_time();
		for (j=0; j<SIZE_INTERVAL_ARRAY; j++)
			fprintf(outfile,"Dist %d used: %d\n", j, dist_trace[j]);
		total_leaf_estimate += leaf_estimate;
		tottime = tottime + time_; 
		i++;
		if (Heuristic_Only) {
			if (tottime >= cpu_seconds)
				more_work = FALSE;
		} else {
			if (i >= paths)
				more_work = FALSE;
		}
	}
	Paths = i;		       /* If we went by time, we'd like     */
				       /* to know the number of iterations  */
	time_ = tottime;                /* for use in add_to_summary() */
	/*for (i=0;i<num;i++) 
		PG_Distribution[i] += hold_pg_ng_dist[i];*/
	leaf_estimate = (total_leaf_estimate / Paths);
	return(leaf_estimate);
}

int kl_pg(int mode, int paths, int print, int threshold)
{
	extern FILE *tfile;
	extern int Heuristic_Only;
	extern int Geometric_Graphs;
	extern int Line_Starts;
	extern int vertex_loc[MAX_VERTICES];
	extern double cpu_seconds;
	extern double time_, tottime;
	extern int alg_type;
	extern int sort_type;
	extern int tbestcost;	/* best KL cut */
	int i=0, low=0, high=num, mid=num/2;
	int kl_cost;
	int more_work = TRUE;
	double leaf_estimate = 0.0;

	int best_of_all;

	sort_type = 7; 		/* always concurrent reordering */

	if (mode != MODE_MAX)
	    printf("error: only maximization available with kl_pg\n"),exit(23);

	alg_type = ALG_TYPE_KL;
	kl_init_loop(DONT_ALLOC);
	if (threshold > 0)  {
		while (more_work) {
			alg_type = ALG_TYPE_KL;
        		if (Geometric_Graphs && Line_Starts)
                		line_randomize();
        		else
                		randomize_vertices();
           		currentcost=compute_initial_cost(node_cost, vertex_loc);
			time_ = log_time();
			perform_maximize(0,0, &kl_cost);
			best_of_all = kl_cost = tbestcost;
			alg_type = BACKTRACKING;
               		init_walk(RESET_LOCS);
                	fix_kl_subset(0, threshold);
                	find_candidates(mode, num_processed);
                	if (explore_one_path(mode,print,&leaf_estimate,
							threshold)==FAIL)
                        	printf("Error - cant use stats and kl-pg\n"), 
							exit(639);
			time_ = log_time();
			tottime += time_;
               		i++;
                	if (Heuristic_Only) {
                       		if (tottime >= cpu_seconds)
                       	         	more_work = FALSE;
                	} else {
                        	if (i >= paths)
                                	more_work = FALSE;
                	}
		}
		return bk_cost;
	} else {
	        kl_main_loop();
        	best_of_all = kl_cost = tbestcost;
        	alg_type = BACKTRACKING;
	}
	while (low <= high) {
		if (mode==MODE_MAX)
			bk_cost = VERY_LOW;
		else
			bk_cost = VERY_HIGH;
		mid = low + (high - low)/2;
		printf("threshold: %d\n", mid);
		greedy_tree_walk(mode, paths, print, mid);
		if (bk_cost < kl_cost)
			low = mid + 1;
		else
			high = mid - 1;
		/***************************this block assumes MAX_CUT********/
		if (bk_cost > best_of_all) {
		   if (tfile != NULL) {
		   fprintf(tfile,"thresh: %d: PG completion beats KL: ", mid);
		   fprintf(tfile,"(best: %d, current: %d, kl: %d)\n",
				 best_of_all, bk_cost, kl_cost);
		   } else {
		   fprintf(stderr,"thresh: %d: PG completion beats KL: ", mid);
		   fprintf(stderr,"(best: %d, current: %d, kl: %d)\n",
				 best_of_all, bk_cost, kl_cost);
		   }
		   best_of_all = bk_cost;
		}
		/*************************************************************/
	}
	if (tfile != NULL)
		fprintf(tfile,"BEST CUT OF ALL: %d\n", best_of_all);
	else 
		fprintf(stderr,"BEST CUT OF ALL: %d\n", best_of_all);
	return(mid);
}

int backtrack(int mode)
{
	extern int Heuristic_Only;
	int i;

	while ( ((i=select(BACKTRACK)) != BOS) && (num_processed < num) )  {
		find_candidates(mode, num_processed);
	}
	if (i != BOS && Augment_With_Local) {
		init_local_opt(vert_loc, num_verts);
		local_opt();
		if (!Heuristic_Only && !check_cost(NULL, currentcost))
			printf("program error 3**********************\n");
	}
	/*if (i != BOS) heap_insert(i);*/
	return(i);
}

/******************************************************************************/
/**  debugging  routines ******************************************************/
/******************************************************************************/

void print_stack()
{
	int i;

	printf("----------------------------------------------\n");
	printf("-- Stack contents-----------------------------\n");
	printf("----------------------------------------------\n");
	for (i=stack_ptr-1; (i>=0); i--) {
		if (stack[i].type == CAND)
			printf("CAND ");
		else if (stack[i].type == UP)
			printf("UP   ");
		printf("%d  ",stack[i].value);
		if (stack[i].side == SET_A)
			printf("SET_A  ");
		else if (stack[i].side == SET_B)
			printf("SET_B  ");
		if (stack[i].greediness == GREEDY)
			printf("GREEDY\n");
		else if (stack[i].greediness == NONGREEDY)
			printf("NONGREEDY\n");
	}
	printf("----------------------------------------------\n");
}

void fprint_stack(char *fname)
{
	FILE *fpp;
	int i;

	if ((fpp=fopen(fname,"w")) == NULL)
		printf("stack file couldn't be opened\n");

	fprintf(fpp,"------------------------------------\n");
	fprintf(fpp,"-- Stack contents-------------------\n");
	fprintf(fpp,"------------------------------------\n");
	for (i=stack_ptr-1; (i>=0); i--) {
		if (stack[i].type == CAND)
			fprintf(fpp,"CAND ");
		else if (stack[i].type == UP)
			fprintf(fpp,"UP   ");
		fprintf(fpp,"%d  ",stack[i].value);
		if (stack[i].side == SET_A)
			fprintf(fpp,"SET_A\n");
		else if (stack[i].side == SET_B)
			fprintf(fpp,"SET_B\n");
	}
	fprintf(fpp,"------------------------------------\n");
	fflush(fpp);
	fclose(fpp);
}

void print_vertices()
{
	int i;

	for (i=0; i<num; i++)
		printf("%d ", vertices[i]);
	printf("\n");
}
