#include <stdio.h> 
#include <math.h>
#include <stdlib.h>
#include "backtrack.h"
#include "part.h"

void free_push_history();
void print_push_history(int *);
int check_cost(int *, int);
void init_backtrack(int);
int backtrack(int);
void nongreedy_stats();
double greedy_tree_walk(int, int, int, int);
FILE *read_sng_file();
int kl_pg(int, int, int, int);
void write_sng_file(int);
void free_geom_nodes();
void dump_cutfiles(int *, int, int);
void print_locs(int *locs);
void kl_init_loop(int);
void kl_main_loop();

/*****************  input data  ******************************************/
extern int num ;              /** number of vertices 			**/
extern double prob;            /** edge probability   			**/
/*************************************************************************/

extern int alg_type ;       	    /** kl or backtracking		  **/
extern int num_verts[], hold_num_verts[] ;
extern int vertices[];
extern int end_of_file,process ;
extern int total_cands;
extern int cand_calls;
extern int big_cands;

extern int tbestcost;			/** set when KL called 	    **/

extern int hold_vertices[MAX_VERTICES]; /** order in which they were pl. **/
extern int vertex_loc[MAX_VERTICES];    /** vertex location (kl)         **/
extern int hold_loc[MAX_VERTICES];      /** temp. storage for best sols  **/
extern int currentcost;   		/** maintain current cost info   **/
extern int output_short;  		/** output flags  **/

extern FILE *outfile;

extern int mode;
extern int bk_cost ;
extern int bk_minsize ;
extern int numedges;
extern int Geometric_Graphs;
extern int Output_P_Count;
extern int Output_Gnuplot;
extern int P_Heuristic;

extern double leaf_count;	    /** leaves in backtracking tree	**/
extern double time_;

extern STK_ELEMPTR stack;

extern int Compare_Partitions;		 /** run KL and compare partitions**/

extern int Paths;		    /** the number of tree walks to       **/
				    /** follow (0-> backtrack thru all)   **/
extern int num_passes;		    /** for kl-augmentation		  **/
extern int KL_PG_Index;
extern double tottime;
extern UP_INFO NG_Up_Array[MAX_VERTS];

int num_graphs =0;


FILE *PG_Distfile;
/**********************************************************/
/**********************************************************/

FILE *read_pg_file()
{
	extern int num_graphs;
	extern int PG_Distribution[];

	int i,j;
	int old_graphs, old_count;
	char fname[80], pr[10], *home;
	FILE *fp;

	home=(char *)malloc(80);
	/*******************************************************************/
	/** for this code to work, V must be a multiple of 100		  **/
	/*******************************************************************/
	if (Geometric_Graphs)
		sprintf(pr, "%.2f.G", prob);
	else
		sprintf(pr, "%.1f", prob);
	home = getenv("HOME");
	sprintf(fname, "%s/research/distributions/pg_dist.%d.%s",home,num,pr+2);
	if ((fp = fopen(fname, "r+")) != NULL) {
		fscanf(fp, "%d", &old_graphs);
		num_graphs += old_graphs;
		for (i=0; i<num; i++) {
			fscanf(fp, "%d %d", &j, &old_count);
			PG_Distribution[i] += old_count;
		}
		fflush(fp);
	} else if ((fp = fopen(fname, "w")) == NULL) 
		fprintf(outfile, "Error in pg_dist file\n"), exit(23);
	rewind(fp);
	return(fp);
}

void write_pg_file(int num_v)
/******************************************************************************/
/** write_pg_file()	The pg_distfile is opened for updating once at	     **/
/**			the beginning of main().  Once open, the number	     **/
/**			of old graphs and the old distributions are read     **/
/**			As new graphs are processed, the file is periodically**/
/**			written out in case the machine dies during a very   **/
/**			long run.  Each time the distribution information is **/
/**			written, a new distribution table (to be put in      **/
/**			tab.c) is written out also.  The file is named:	     **/
/**			pg_dist.<num>.<prob> and located in ~/research       **/
/******************************************************************************/
{
	extern int PG_Distribution[];
	int i,j,k;
	int step;				/* how many vertices make  */
						/* up 1% of the vertices   */
						/* (assume >=100 verts.)   */
	double d_i_total;			/* sum of d(i) values over */
						/* an interval representing*/
						/* 1% of the placements.   */
	double	dist_table[DISTRIB_SIZE];	/* lookup table for prob.  */
						/* heuristic.              */
						/* after i% of vertices    */
						/* have been placed, put   */
						/* nongreedy with prob dt[i]*/
						/* ONLY WORKS IF 100 | V    */

	/*******************************************************************/
	/** for this code to work, V must be a multiple of 100		  **/
	/*******************************************************************/
	fprintf(PG_Distfile,"%d \n",num_graphs);
	for (i=0; i<num_v; i++) /* assume num same for all G */
	    fprintf(PG_Distfile,"%d %d\n",i,PG_Distribution[i]);
	step = (int) (num_v * 0.01);
	fprintf(PG_Distfile,"LOOKUP_TABLE\n");
	fprintf(PG_Distfile,"------------\n");
	for (i=0, k=0; i<num_v; i=i+step, k=k+1) {
		d_i_total = 0.0;
		for (j=i; j<i+step; j++)
			d_i_total = d_i_total +
 		         (PG_Distribution[j]/(double)num_graphs);
		dist_table[k]= d_i_total / (double)step;
	        fprintf(PG_Distfile,"%d %.6f\n", k, dist_table[k]);
	}
	fflush(PG_Distfile);
	rewind(PG_Distfile);
}

int main(int argc, char **argv)
{
	extern int Output_Sng_Distfile;
	extern int KL_PG_Flag;
	extern int KL_PG_Alg;
	extern int KL_PG_Threshold;
	extern int Nongreedy_Stats;
	extern int Heuristic_Only;
	extern int Print_Cand_History;
	extern int max_part_diff;
	extern int PG_Distribution[];
	extern double total_part_diff;
	extern FILE *Sng_Distfile;
	int i, partition_diff, first = TRUE;
	int backup_num=0;
	static char hyphens[] = "--------------------------------------------------------------------------------";

	alg_type = BACKTRACKING;
	num_graphs = 0;
	clear_summary();
	process_parameters(argc, argv);
	end_of_file = FALSE;
	for (i=0; i<num; i++) PG_Distribution[i] = 0;
        if (Output_Sng_Distfile)
                init_extra_file_info();
	while (!end_of_file) {
		if (!read_data()) {
			printf("invalid data input, program terminated...\n");
			exit(1);
		}
		if (first) {
			if (Output_P_Count) 
				PG_Distfile = read_pg_file();
                        if (Output_Sng_Distfile) 
                                Sng_Distfile = read_sng_file();
			first = FALSE;
		}
		if (process) {
			init_backtrack(CREATE_GRAPH);
			/*printf("Paths:%d, (Paths==BACKTRACK):%d\n",Paths,
					(Paths == BACKTRACK));*/

			if (Paths == BACKTRACK) {
				time_ = log_time();
               			while((i=backtrack(mode)) != BOS); 
				time_ = log_time();
				if (Print_Cand_History) {
					print_push_history(hold_vertices);
					free_push_history();
				}
			} else if (!KL_PG_Flag) {
				leaf_count=
				greedy_tree_walk(mode, Paths,PRINT, 0);
			} else {
				if (KL_PG_Alg)
				  KL_PG_Index = kl_pg(mode, Paths, PRINT,
					KL_PG_Threshold);
				else
				  KL_PG_Index = kl_pg(mode, Paths, PRINT,FALSE);
			}
			if (!Heuristic_Only && !check_cost(hold_loc, bk_cost))
				printf("*************** program error 5\n");

            		if(num <= 10) 
               			print_locs(hold_loc);
			currentcost = bk_cost;
               		if (Nongreedy_Stats)
                       		 nongreedy_stats(); 
            		add_to_summary();
         		if(!output_short && Paths == BACKTRACK) {
            			fprintf(outfile,"%66s\n", hyphens);
               			fprintf(outfile,"%-8s%-6s%-10s%-9s%-9s%-7s%-7s%-6s%-4s\n", 
				"TIME", "CUTS",  "AVE.CANDS", "LEAVES",
				"TIMES>=2", "M.P.D.", "A.P.D.","SET_A","SET_B");
            			fprintf(outfile,"%66s\n", hyphens);
               		        fprintf(outfile,"%-8.2f%-6d%-11.3f%-9.0f%-9d%-7d%-7.2f%-6d%-4d\n",
                     		time_,bk_cost, 
				(double) total_cands / (double) cand_calls,
				leaf_count,
				/*pow(10.0,(double)log10((double)leaf_count)/num),*/
				big_cands,
				max_part_diff,
				total_part_diff / leaf_count,
                     		hold_num_verts[SET_A],hold_num_verts[SET_B]);
			}

         		if(output_short)
            			fprintf(outfile,"Cuts in best partition: %5d\n",
								bk_cost);
         		else {
				char   solution[20] = "                   ";

				if (MIN_QUOTIENT)
					sprintf(solution,"%10.6f (%d)",
					   (double) bk_cost/(double) bk_minsize,
                                  		bk_cost);
				else
					sprintf(solution,"%-20d", bk_cost);
		    	  	fprintf(outfile,"\n TOTAL->-> %6.2f%-20s\n",
			  	tottime,solution);
			}
			if (Output_Gnuplot)
				dump_cutfiles(vertices, 0, num), exit(1);
			if (KL_PG_Flag)
			   fprintf(outfile,"KL start/PG finish threshold: %d\n",
					KL_PG_Index);
         		fflush(outfile);
         		free_adj_list() ;
			free(stack);
			free_buckets();
			if (Geometric_Graphs)
				free_geom_nodes();
			if (Compare_Partitions) {
				alg_type = ALG_TYPE_KL;
				partition_diff = num;
				if (!Heuristic_Only)
					num_passes = Paths;
				kl_init_loop(ALLOC);
				kl_main_loop();
	 	     		fprintf(outfile,"\n TOTAL->-> %6.2f  %5d\n",
		     		tottime,tbestcost);
				for (i=0; i<num; i++)
				   partition_diff-=(vertex_loc[i]==hold_loc[i]);
				fprintf(outfile,"PARTITION DIFFERENCE:\n");
				fprintf(outfile,"%d same, %d different\n",
					num-partition_diff,partition_diff);
				alg_type = BACKTRACKING;
			}
			backup_num = num;
			num_graphs++;
               		if (num_graphs % 10 == 0) {
                       		 if (Output_Sng_Distfile)
                                	write_sng_file(num);
				if (Output_P_Count)
					write_pg_file(num);
			}

       		}
    	}
        if (Output_Sng_Distfile)
        	write_sng_file(backup_num);
        if (Output_P_Count)
        	write_pg_file(backup_num);
   	num = 0; prob = 0.0;
   	print_summary();
	
	/********* debug ********
	{
		extern int P_Interval;
		extern PG_DIST ng_distrib[][DISTRIB_SIZE];
		double temp= 0;
		int j;

		for (i=0; i<3; i++) {
			for (j=0; j<DISTRIB_SIZE; j++)
				temp += (ng_distrib[i][j] * P_Interval);
			printf("%.1f ", temp);
			temp = 0.0;
		}
	}
	********* debug ********/
	exit(0);       			/** useful in shell scripting **/
}
