#include <stdio.h>
#include "part.h" 

extern int 	num;
extern int	num_vertices[2];

int	Current_Max_Bucket[2];
int	Current_Min_Bucket[2];
int	Num_Pos_Buckets;
int	Num_Buckets;

LISTPTR *buckets[2], dummies, nodes;
	

void alloc_buckets(int num_buckets)
/*************************************************************************/
/****************************** alloc_buckets() **************************/
/*************************************************************************/
/** DESCRIP:  	An array of buckets is allocated.  The number of 	**/
/**		buckets is determined to allow pos & neg values up 	**/
/**		to the maximum difference in cost for a node (num-1)    **/
/** PARAMS:	num_buckets : the TOTAL number of buckets, pos & neg	**/
/*************************************************************************/
{
	int i;

	Num_Buckets = num_buckets;
	Num_Pos_Buckets = (num_buckets-1)/2;
	buckets[0] = (LISTPTR *) calloc(num_buckets, sizeof(LISTPTR));
	buckets[1] = (LISTPTR *) calloc(num_buckets, sizeof(LISTPTR));
	dummies = (LISTPTR) calloc(2*num_buckets, sizeof(LIST));
	for (i=0; i<num_buckets; i++) {
		buckets[0][i] = &dummies[i];
		buckets[0][i]->prev  = &dummies[i];
		buckets[1][i] = &dummies[i+num_buckets];
		buckets[1][i]->prev  = &dummies[i+num_buckets];
	}
	nodes = (LISTPTR) calloc(num, sizeof(LIST));
	for (i=0; i<num; i++) 
		nodes[i].fix_next = FALSE;
}

void free_buckets()
/*************************************************************************/
/****************************** free_buckets() ***************************/
/*************************************************************************/
{
	free((char *) buckets[0]);
	free((char *) buckets[1]);
	free((char *) dummies);
	free((char *) nodes);
}

void init_buckets(int reset_locs)
/*************************************************************************/
/** init_buckets() 	: only for backtracking alg **********************/
/*************************************************************************/
{
	int i;

	Current_Max_Bucket[0] = Current_Min_Bucket[0] = Num_Pos_Buckets;
	Current_Max_Bucket[1] = Current_Min_Bucket[1] = Num_Pos_Buckets;
	for (i=0; i<Num_Buckets; i++) {
		buckets[0][i]->prev  = &dummies[i];
		buckets[1][i]->prev  = &dummies[i+Num_Buckets];
		buckets[0][i]->next  = NULL;
		buckets[1][i]->next  = NULL;
	}
	for (i=0; i<num; i++) {
		nodes[i].node_no = i;
		if (reset_locs)
			nodes[i].loc = 0;
		nodes[i].lock = NOT_LOCKED;
		nodes[i].cost = 0;
		insert_to_bucket(0, nodes[i].cost, &nodes[i]); 
	}
}
/*************************************************************************/

void reset_buckets(int (*costs)[MAX_VERTICES])
/*************************************************************************/
/***************************** reset_buckets() ***************************/
/*************************************************************************/
/** DESCRIP: The buckets arrays are initialized such that each bucket   **/
/**		points to a dummy node.  Then the nodes are distributed **/
/**		to the buckets according to their loc fields and the	**/
/**		costs array.						**/
/** PARAMS: 	costs: cost for each node to be in each partition	**/
/*************************************************************************/
{
	int i, loc;

	Current_Max_Bucket[0] = Current_Min_Bucket[0] = Num_Pos_Buckets;
	Current_Max_Bucket[1] = Current_Min_Bucket[1] = Num_Pos_Buckets;
	for (i=0; i<Num_Buckets; i++) {
		buckets[0][i]->prev  = &dummies[i];
		buckets[1][i]->prev  = &dummies[i+Num_Buckets];
		buckets[0][i]->next  = NULL;
		buckets[1][i]->next  = NULL;
	}
	for (i=0; i<num; i++) {
		loc = nodes[i].loc;
		nodes[i].node_no = i;
		nodes[i].lock = NOT_LOCKED;
		nodes[i].cost = costs[!loc][i] - costs[loc][i];
		if (!nodes[i].fix_next)			/** 3/20/93 **/
			insert_to_bucket(loc, nodes[i].cost, &nodes[i]); 
		else {
			nodes[i].fix_next = 0;
			nodes[i].lock = LOCKED;
		}
	}
}

void insert_to_bucket(int side, int cost, LISTPTR node)
/*************************************************************************/
/***************************** insert_to_bucket() ************************/
/*************************************************************************/
/** DESCRIP: Insert <node> into the bucket in the <side> buckets array  **/
/**	     at position <cost>.  Note that if <cost> is 0, the insert  **/
/** 	     occurs at the <num-1>th bucket.  If the <cost> is the most **/
/**	     negative possible, the insert occurs at bucket 0.		**/
/**	     Insertion occurs at the end of the bucket list, and the	**/
/**	     <prev> field of the dummy element of the list serves as an **/
/**	     end_of_list pointer.					**/
/*************************************************************************/
{
	int i = cost + Num_Pos_Buckets;

	/*if (!node)	printf("Error: insert_to_bucket()\n"), exit(3);*/

	node->loc = side;
	node->cost = cost;
	node->bucket_no = i;

	node->next = NULL;
	node->prev = buckets[side][i]->prev;
	buckets[side][i]->prev->next = node;
	buckets[side][i]->prev = node;

	if (i > Current_Max_Bucket[side]) {
		Current_Max_Bucket[side] = i;
	} else if (i < Current_Min_Bucket[side])
		Current_Min_Bucket[side] = i;
}

void delete_from_bucket(LISTPTR node)
/*************************************************************************/
/***************************** delete_from_bucket() **********************/
/*************************************************************************/
/** DESCRIP:  	This routine is the reason that doubly linked lists	**/
/**		are necessary.  Given a pointer to a node, delete it	**/
/**		from its bucket list.  Note that the Current_Max and	**/
/**		Current_Min indexes may be incorrect after this 	**/
/**		routine.  However, they are always correct bounds.	**/
/**		The extra work of actually finding the min and max is	**/
/**		left for the extract routines since they are called so	**/
/**		much less frequently.					**/
/** PARAMS:	node : ptr to the node to be deleted			**/
/*************************************************************************/
{
	int 	i, side=node->loc;

	i = node->bucket_no;

	node->prev->next = node->next;
	if (node->next)
		node->next->prev = node->prev;
	if (buckets[side][i]->prev == node)
		buckets[side][i]->prev = node->prev;

	if (buckets[side][i]->next==NULL) {
		if (i==Current_Max_Bucket[side])
			Current_Max_Bucket[side] = i-1;
		else if (i==Current_Min_Bucket[side])
			Current_Min_Bucket[side] = i+1;
	}
}

void move_to_bucket(int side, int dest, LISTPTR node)
/*************************************************************************/
/***************************** move_to_bucket() **************************/
/*************************************************************************/
/** DESCRIP:  	Change a node's location in the bucket arrays.  This 	**/
/**		is called when one of the node's neighbors moves from	**/
/**		one partition of the graph to the other.		**/
/*************************************************************************/
{
	delete_from_bucket(node);
	insert_to_bucket(side, dest, node);
}

LISTPTR bucket_extract_max(int lock)
/*************************************************************************/
/***************************** bucket_extract_max() **********************/
/*************************************************************************/
/** DESCRIP: 	Choose the next best cost node to move from one partition*/
/**		to the other.  Update the Current_Max_Bucket[] values	**/
/** 		to reflect the actually maxima.  			**/
/** PARAMS:	lock: LOCKED if kl called this, NOT_LOCKED if local_opt **/
/*************************************************************************/
{
	int i;
	LISTPTR temp=NULL;
	
	if (buckets[0][Current_Max_Bucket[0]]->next == NULL) {
		for (i=Current_Max_Bucket[0]; i >= Current_Min_Bucket[0]; i--)
			if (buckets[0][i]->next != NULL) 
				break;
		Current_Max_Bucket[0] = (i >= 0) ? i : 0;
	}
	
	if (buckets[1][Current_Max_Bucket[1]]->next == NULL) {
		for (i=Current_Max_Bucket[1]; i >= Current_Min_Bucket[1]; i--)
			if (buckets[1][i]->next != NULL) 
				break;
		Current_Max_Bucket[1] = (i >= 0) ? i : 0;
	}
	
	if (Current_Max_Bucket[0] > Current_Max_Bucket[1])  {
		if (buckets[0][Current_Max_Bucket[0]]->next != NULL) 
			temp = buckets[0][Current_Max_Bucket[0]]->next;
		else if (buckets[1][Current_Max_Bucket[1]]->next != NULL)
			temp = buckets[1][Current_Max_Bucket[1]]->next;
		else
			printf("Error: extract_max\n");
	} else {
		if (buckets[1][Current_Max_Bucket[1]]->next != NULL) 
			temp = buckets[1][Current_Max_Bucket[1]]->next;
		else if (buckets[0][Current_Max_Bucket[0]]->next != NULL)
			temp = buckets[0][Current_Max_Bucket[0]]->next;
		else
			printf("Error: extract_max\n");
	}
	temp->lock = lock;
	delete_from_bucket(temp);
	return(temp);
}

LISTPTR bucket_extract_min(int lock)
/*************************************************************************/
/***************************** bucket_extract_min() **********************/
/*************************************************************************/
/** DESCRIP: 	Choose the next best cost node to move from one partition*/
/**		to the other.  Since  we are in minimize mode, make sure**/
/**		that the balance criteria are satisfied.		**/
/**		Update the Current_Min_Bucket[] values			**/
/** PARAMS:	lock: LOCKED if kl called this, NOT_LOCKED if local_opt **/
/*************************************************************************/
{
	extern int currentcost;
	extern int mode;
	int i, new_cost0, new_cost1;
	int    new_size0, new_size1;
	double new_quot0, new_quot1;
	LISTPTR best_a, best_b, temp=NULL;
	
	if (buckets[0][Current_Min_Bucket[0]]->next == NULL) {
		for (i=Current_Min_Bucket[0]; i <= Current_Max_Bucket[0]; i++)
			if (buckets[0][i]->next != NULL) 
				break;
		Current_Min_Bucket[0] = (i<Num_Buckets) ? i : Num_Buckets-1;
	}
	
	if (buckets[1][Current_Min_Bucket[1]]->next == NULL) {
		for (i=Current_Min_Bucket[1]; i <= Current_Max_Bucket[1]; i++)
			if (buckets[1][i]->next != NULL) 
				break;
		Current_Min_Bucket[1] = (i<Num_Buckets) ? i : Num_Buckets-1;
	}
	best_a = buckets[0][Current_Min_Bucket[0]]->next;
	best_b = buckets[1][Current_Min_Bucket[1]]->next;
	if (mode == MIN_QUOTIENT) {
	   if ((best_a != NULL) && (best_b != NULL)) {
	   	new_cost0 = currentcost + best_a->cost;
	   	new_cost1 = currentcost + best_b->cost;
                new_size0 = MIN(num_vertices[0]-1, num_vertices[1]+1);
                new_size1 = MIN(num_vertices[1]-1, num_vertices[0]+1);
	   	if (new_size0 > 0) 
			new_quot0 = ((double) new_cost0 / (double) new_size0);
		else
			new_quot0 = MAX_INT;
	   	if (new_size1 > 0) 
			new_quot1 = ((double) new_cost1 / (double) new_size1); 
		else
			new_quot1 = MAX_INT;

	   	if (new_quot0 < new_quot1)
			temp = best_a;
		else
			temp = best_b;
	   } else if (best_a != NULL) {
			temp = best_a;
	   } else if (best_b != NULL) {
			temp = best_b;
	   } else 	printf("error: bucket_extract_min(): empty buckets\n");
	} else {
		if (Current_Min_Bucket[0] < Current_Min_Bucket[1]) {
			if ((num_vertices[1] - num_vertices[0] < UNBALANCE) &&
			    (best_a != NULL) )
				temp = best_a;
			else
				temp = best_b;
		} else {
			if ((num_vertices[0] - num_vertices[1] < UNBALANCE) &&
		    	(best_b != NULL) )
				temp = best_b;
			else
				temp = best_a;
		}
	}
	temp->lock = lock;
	delete_from_bucket(temp);
	return(temp);
}

void print_buckets()
{
	int i;
	LISTPTR temp;

	printf("***--------------------------------------------\n");
	printf("***---------------BUCKETS[0]-------------------\n");
	printf("***--------------------------------------------\n");
	for (i=Current_Max_Bucket[0]; i>=Current_Min_Bucket[0]; i--) {
		printf("*** %d: ",i-Num_Pos_Buckets);
		for (temp=buckets[0][i]->next; temp; temp = temp->next)
				printf(" %d", temp->node_no);
		printf("\n");
	}
	printf("***--------------------------------------------\n");
	printf("***---------------BUCKETS[1]-------------------\n");
	printf("***--------------------------------------------\n");
	for (i=Current_Max_Bucket[1]; i>=Current_Min_Bucket[1]; i--) {
		printf("*** %d: ",i-Num_Pos_Buckets);
		for (temp=buckets[1][i]->next; temp; temp = temp->next)
				printf(" %d", temp->node_no);
		printf("\n");
	}
	printf("***--------------------------------------------\n");
}

void print_buckets1()
{
	int i;
	LISTPTR temp;

	printf("***--------------------------------------------\n");
	printf("***---------------BUCKETS[0]-------------------\n");
	printf("***--------------------------------------------\n");
	for (i=Current_Max_Bucket[0]; i>=Current_Min_Bucket[0]; i--) {
		printf("*** %d: ",i-Num_Pos_Buckets);
		for (temp=buckets[0][i]->next; temp!=buckets[0][i]->prev; 
						temp = temp->next)
			printf(" %d", temp->node_no);
		printf(" %d", temp->node_no);
		printf("\n");
	}
	printf("***--------------------------------------------\n");
	printf("***---------------BUCKETS[1]-------------------\n");
	printf("***--------------------------------------------\n");
	for (i=Current_Max_Bucket[1]; i>=Current_Min_Bucket[1]; i--) {
		printf("*** %d: ",i-Num_Pos_Buckets);
		for (temp=buckets[1][i]->next; temp!=buckets[1][i]->prev; 
						temp = temp->next)
			printf(" %d", temp->node_no);
		printf(" %d", temp->node_no);
	}
	printf("***--------------------------------------------\n");
}

void print_bounds()
{
	int i;

	printf("***--------------------------------------------\n");
	printf("***---------------BUCKETS[0]-------------------\n");
	printf("***--------------------------------------------\n");
	for (i=Current_Max_Bucket[0]; i>=Current_Min_Bucket[0]; i--) {
		printf("*** %d: ",i-Num_Pos_Buckets);
		if (buckets[0][i]->next)
			printf(" %d ", buckets[0][i]->next->node_no);
		else    printf(" *");
		if (buckets[0][i]->prev != buckets[0][i])
	 		printf("%d\n" , buckets[0][i]->prev->node_no);
		else    printf(" *\n");
	}
	printf("***--------------------------------------------\n");
	printf("***---------------BUCKETS[1]-------------------\n");
	printf("***--------------------------------------------\n");
	for (i=Current_Max_Bucket[1]; i>=Current_Min_Bucket[1]; i--) {
		printf("*** %d: ",i-Num_Pos_Buckets);
		if (buckets[1][i]->next)
			printf(" %d ", buckets[1][i]->next->node_no);
		else    printf(" *");
		if (buckets[1][i]->prev != buckets[1][i])
	 		printf("%d\n" , buckets[1][i]->prev->node_no);
		else    printf(" *\n");
	}
	printf("***--------------------------------------------\n");
}

int bucket_cost()
{
	int i, bcost=0;
	LISTPTR temp;

	for (i=Current_Max_Bucket[0]; i>=Current_Min_Bucket[0]; i--) 
		for (temp=buckets[0][i]->next; temp; temp = temp->next)
			bcost += temp->cost;
	for (i=Current_Max_Bucket[1]; i>=Current_Min_Bucket[1]; i--) 
		for (temp=buckets[1][i]->next; temp; temp = temp->next)
			bcost += temp->cost;
	return bcost;
}

LISTPTR extract_min_from_side(int side, int lock)
/*************************************************************************/
/***************************** bucket_extract_min() **********************/
/*************************************************************************/
{
	int i;
	LISTPTR temp=NULL;
	
	if (buckets[side][Current_Min_Bucket[side]]->next == NULL) {
		for (i=Current_Min_Bucket[side];i<=Current_Max_Bucket[side];i++)
			if (buckets[side][i]->next != NULL) 
				break;
		Current_Min_Bucket[side] = i;
	}
	if (buckets[side][Current_Min_Bucket[side]]->next != NULL) 
		temp = buckets[side][Current_Min_Bucket[side]]->next;
	else {
		printf("Error: extract_max\n");
		exit(759);
	}
	temp->lock = lock;
	delete_from_bucket(temp);
	return(temp);
}
