#include <stdio.h>
#include <math.h>
#include "part.h" 

#define	NUM_BUCKETS 100		/* must be a mult of 100	*/

double rrandom();

typedef struct point *POINTPTR;

typedef struct point {
	int 	num;		/* vertex number 	*/
	double  y_int;		/* y intercept 		*/
	double 	x;		/* x coordinate		*/
	double 	y;		/* y coordinate		*/
	POINTPTR next;
} POINT;

extern int num;

int		Current_Bucket;
POINTPTR	Current_Node;

POINTPTR g_buckets[NUM_BUCKETS], node_array;
POINT    dumb[NUM_BUCKETS];
	
void insert_in_order(POINTPTR);

void print_coordinates()
{
	extern FILE *rcfile;

	int i;
	for (i=0; i<num; i++) 
	   fprintf(rcfile,"%8.6f %8.6f\n", node_array[i].x,node_array[i].y);
	fclose(rcfile);
}

void print_node_array()
{
	int i;
	for (i=0; i<num; i++) 
		printf("(%6d%6.2f%6.2f%6.2f)\n", node_array[i].num,
			node_array[i].y_int, node_array[i].x,
			node_array[i].y);
}

void free_geom_nodes()
/*************************************************************************/
/****************************** free_geom_nodes() ************************/
/*************************************************************************/
{
	free((char *) node_array);
}

void init_geom_buckets()
/*************************************************************************/
/****************************** init_geom_buckets() **********************/
/*************************************************************************/
{
	int i;

	Current_Bucket = -1;
	/* Current_Bucket = 0;		*/ /* until 3/94 */
	Current_Node   = NULL;
	for (i=0; i<NUM_BUCKETS; i++) {
		g_buckets[i]  = &dumb[i];
		dumb[i].y = -1.0;
		dumb[i].next = NULL;
	}
	node_array = (POINTPTR) calloc(num, sizeof(POINT));
	for (i=0; i<num; i++) {
		node_array[i].num = i;
		node_array[i].x = rrandom();	/* 4/27/93 */
		node_array[i].y = rrandom();	/* 4/27/93 */
		node_array[i].next = NULL;
		insert_in_order(&node_array[i]); 
	}
}

int y_int_compare(POINTPTR p1, POINTPTR p2)
{
	if ((p1->y_int - p2->y_int) > 0)
		return 1;
	else if ((p1->y_int - p2->y_int) == 0)
		return 0;
	else
		return -1;
}

void line_randomize()
{
	extern double rrandom2();
	extern LISTPTR nodes;
	extern int Starting_Locs[MAX_VERTICES]; /* for Simulated Annealing */
	extern int num_vertices[2];
	extern int vertex_loc[MAX_VERTICES];

	int i;
	double radians = rrandom2() * M_PI;
	double tan_theta = tan(radians);

	for (i=0; i<num; i++)
		node_array[i].y_int = node_array[i].y - 
					(tan_theta * node_array[i].x);
	qsort(node_array, num, sizeof(POINT),  
			(int (*)(const void *, const void *)) y_int_compare);
	for (i=0; i<num/2; i++) {
		vertex_loc[node_array[i].num] = SET_A;
		nodes[node_array[i].num].loc = SET_A;
		Starting_Locs[node_array[i].num] = SET_A;
	}
	for (i=(num/2)+1; i<num; i++) {
		vertex_loc[node_array[i].num] = SET_B;
		nodes[node_array[i].num].loc = SET_B;
		Starting_Locs[node_array[i].num] = SET_B;
	}
	num_vertices[SET_A] = num/2;
	num_vertices[SET_B] = num - num/2;
}
	

void insert_in_order(POINTPTR node)
/*************************************************************************/
/***************************** insert_in_order() *************************/
/*************************************************************************/
/** DESCRIP: Insert <node> into the bucket sorted by y coordinate       **/
/**          insert in order so that the bucket lists are sorted.	**/
/*************************************************************************/
{
	int b;
	POINTPTR temp;

	b = (int)  (node->y * NUM_BUCKETS);
	for (temp=g_buckets[b];temp->next&&(temp->y < node->y);temp=temp->next);
	node->next = temp->next;
	temp->next = node;
}

void findnext()
{
	extern int 	Current_Bucket;
	extern POINTPTR Current_Node;

	if (Current_Node  && Current_Node->next)
		Current_Node = Current_Node->next;
	else 
		for (;++Current_Bucket < NUM_BUCKETS;)  {
			if (g_buckets[Current_Bucket]->next != NULL) {
				Current_Node = g_buckets[Current_Bucket]->next;
				break;
			}
		}
}

double dist(POINTPTR node1, POINTPTR node2)
{
	double ydiff, xdiff, d;
	
	ydiff = node1->y - node2->y;
	xdiff = node1->x - node2->x;

	d = sqrt(ydiff*ydiff + xdiff*xdiff);
	return(d);
}

void check_in_list(POINTPTR node, double d)
{
	extern int numedges; /* , deg[];*/
	POINTPTR temp;
	double distance1;

	if (node == NULL) return;

	for (temp=node; temp; temp=temp->next) {
		distance1 = dist(Current_Node, temp);
		if (distance1 < d) {
			add_adj_entry(Current_Node->num, temp->num);
			add_adj_entry(temp->num, Current_Node->num);
			numedges++;
			/*deg[Current_Node->num]++;
			deg[temp->num]++;*/
		}
	}
}
	



void gen_geom_graph(int n, double d)
{
	extern POINTPTR Current_Node;

	int i, j, incr;
	POINTPTR new_node;

	for (i=0; i<n-1; i++) {
		findnext();
		check_in_list(Current_Node->next, d);
		incr = (int) (NUM_BUCKETS*d) + 1;
		for (j=Current_Bucket+1; j<=(Current_Bucket+incr);j++) {
			if (j>=NUM_BUCKETS) break;
			if (g_buckets[j]->next==NULL) continue;
			new_node = g_buckets[j]->next;
			check_in_list(new_node, d);
		}
	}
}

/*
void gen_geom_graph(int n, double d)
{
	extern int numedges;
	int i, j;
	for(i=0;i<num;i++) {
		for(j=0;j<num;j++) {
			if((i!=j) && dist(&node_array[i], &node_array[j]) < d) {
				add_adj_entry(i,j);
				numedges++;
			}
		}
	}
	numedges = numedges / 2;
}
*/
void print_geom_buckets()
{
	int i;
	POINTPTR temp;

	printf("***--------------------------------------------\n");
	printf("***---------------BUCKETS----------------------\n");
	printf("***--------------------------------------------\n");
	for (i=NUM_BUCKETS-1; i>=0; i--) {
		printf("*** %d: ",i);
		for (temp=g_buckets[i]->next; temp; temp = temp->next)
				printf(" [%d, (%.4f,%.4f)]", temp->num,
					temp->x, temp->y);
		printf("\n");
	}
	printf("***--------------------------------------------\n");
}


void create_geom_graph(double distance)
{
	extern int num;
	extern int Output_Rao;
	extern int Line_Starts;

	/*fgets(str, 80, stdin);
	sscanf(str, " %d %.6lf\n", &num, &d);*/
	
	init_geom_buckets();
	gen_geom_graph(num,distance);
	if (Output_Rao && Line_Starts)
		print_coordinates();
	/*print_geom_buckets();*/
	/*free_geom_nodes();*/	/* freed later */
}

int num_compare(POINTPTR p1, POINTPTR p2)
{
	return p1->num - p2->num;
}


void dump_points()
{
	int i;
	double x,y;
	FILE *fp1;
	
	if ((fp1=fopen("points","w")) == NULL)
		printf("Error: can't open cut file\n");

	for (i=0; i<num; i++) {
		x = node_array[i].x;
		y = node_array[i].y;
		fprintf(fp1,"%f\t%f\n", x, y);
	}
}

void dump_cutfiles(int *verts, int lower_limit, int upper_limit)
{
	extern char *scan_string();
	extern int alg_type;
	extern int Geometric_Graphs;
	extern int hold_loc[MAX_VERTICES];
	extern int post_tots[MAX_VERTICES];
	extern LISTPTR nodes;
	int i, j, loc1, loc2, v, v2, edges;
	int tot_deg = 0;
	double x,y, x2, y2;
	FILE *fp1, *fp2, *fp3, *fp4, *fp5;
	
	if (!Geometric_Graphs)
		init_geom_buckets();
	
	qsort(node_array, num, sizeof(POINT), 
			(int (*)(const void *, const void *)) num_compare);
	if ((fp1=fopen("cuts","w")) == NULL)
		printf("Error: can't open cut file\n");
	
	if ((fp2=fopen("a_edges","w")) == NULL)
		printf("Error: can't open a_edges file\n");
	
	if ((fp3=fopen("b_edges","w")) == NULL)
		printf("Error: can't open b_edges file\n");
	
	if ((fp4=fopen("gr_verts","w")) == NULL)
		printf("Error: can't open greedy file\n");
	
	if ((fp5=fopen("ng_verts","w")) == NULL)
		printf("Error: can't open nongreedy file\n");

	for (i=lower_limit; i<upper_limit; i++) {
	   v = verts[i];
	   x = node_array[v].x;
	   y = node_array[v].y;
	   edges = 0;
	   for (j=i; j<upper_limit; j++) {
	   	v2 = verts[j];
		if (adj(v, v2)) {
	   		x2 = node_array[v2].x;
	   		y2 = node_array[v2].y;
			if (alg_type == ALG_TYPE_KL) {
				loc1 = nodes[v].loc;
				loc2 = nodes[v2].loc;
			} else {
				loc1 = hold_loc[v];
				loc2 = hold_loc[v2];
			}
	   		/*if (post_tots[lower_limit+i] < 0) {*/
				if (loc1 != loc2) {
					fprintf(fp1,"%f\t%f\n", x, y);
					fprintf(fp1,"%f\t%f\n\n",x2, y2);
				} else if (loc1 == 0) {
					fprintf(fp2,"%f\t%f\n", x, y);
					fprintf(fp2,"%f\t%f\n\n", x2, y2);
				} else {
					fprintf(fp3,"%f\t%f\n", x, y);
					fprintf(fp3,"%f\t%f\n\n", x2, y2);
				}
			/*}*/
			edges ++;
	   	}
	   }
	   tot_deg += (2*edges);
	   if (post_tots[lower_limit+i] >= 0)
		/*fprintf(fp4,"%f\t%f\t%d\t%d\n", x, y, v, edges);*/;
	   else
		fprintf(fp5,"%f\t%f\t%d\t%d\n", x, y, v, edges);
	}
	printf("Average Sg Degree: %f\n", tot_deg / 
					(double) (upper_limit-lower_limit));
	fclose(fp1);
	fclose(fp2);
	fclose(fp3);
	fclose(fp4);
	fclose(fp5);
	if (!Geometric_Graphs)
		free_geom_nodes();
}
