/**************************************************************************/
/****************************** lex.l *************************************/
/**************************************************************************/
%{
#ifdef LEXDEBUG

#include <assert.h>
#include <floatingpoint.h>

int main()
{
	char *p;
	int i;

	assert(sizeof(int) >= sizeof(char *));
	while (p = (char *) yylex()) 
		printf("token %s is \"%s\"\n",p, yytext);
}
#define token(x)	(int) "x"

#else

#define token(x)	x

#endif

#include<malloc.h>
#include<stdlib.h>
#include<math.h>
#define INT	250
#define FLOAT	251
#define STRING	252

%}

letter 		[a-zA-Z_]
digit		[0-9]
letter_or_digit	[a-zA-Z_0-9]
white		[ \t\n]
blank		[ \t]
other		.
dot		\.

%%

{digit}+						{ return token(INT);  }
{digit}*{dot}{digit}+("E"[+-]?{digit}+)?		{ return token(FLOAT);}

{letter}{letter_or_digit}*				{ return token(STRING);}
{white}+						;
{other}							;

%%

int scan_int(int lowerbound, int upperbound)
{
	int t, i = lowerbound-1;

	while (i<lowerbound || i>upperbound) {
		while ((t=yylex()) != INT) {
			printf("Expecting integer\n");
			if (t == 0) {
				printf("Unexpected EOF\n");
				exit(1);
			}
		}
		i = atoi(yytext);
		if (i<lowerbound || i>upperbound)
			printf("not in range [%d...%d]\n",lowerbound,
							    upperbound);
	}
	return(i);
}

double scan_double(double lowerbound, double upperbound)
{
	int t;
	double f = lowerbound-1;

	while (f<lowerbound || f>upperbound) {
		while ((t=yylex()) != FLOAT) {
			printf("Expecting double\n");
			if (t == 0) {
				printf("Unexpected EOF\n");
				exit(1);
			}
		}
		f = (double) atof(yytext);
		if (f<lowerbound || f>upperbound)
			printf("not in range [%.6f...%.6f]\n",lowerbound,
							    upperbound);
	}
	return((double) f);
}

char * scan_string()
{
	int t, i=0;
	char *temp, *p;

	while ((t=yylex()) != STRING)
		printf("Expecting character string\n");
		if (t == 0) {
			printf("Unexpected EOF\n");
			exit(1);
		}
	temp = (char * ) malloc(80);
	p = temp;
	while ((*temp++ = yytext[i++]));
	return(p);
}

#ifdef LEXDEBUG
FILE *tfile;
#endif

int yywrap() 
{ 
	extern FILE *tfile; 
	if (tfile)  
		fprintf(tfile,"ERROR: yywrap() called with dummy definition\n");
	else
		printf("ERROR: yywrap() called with dummy definition\n");
	return 1;
}

/*
main()
{
	int i;
	double f;
	char *s, fname1[10], fname2[10];
	FILE *fp;
	
	printf("Enter filename: ");
	scanf("%s", fname1);	
	if (strcmp(fname1,"stdin")!=0 && (fp=freopen(fname1, "r", stdin)) != 0)
		printf("file error\n");


	printf("Enter integer: ");
	printf("\nentered: %d\n",  scan_int(0,22));
	printf("Enter double: ");
	printf("\nentered: %.6f\n", scan_double(0.0,1.0));
	printf("Enter string: ");
	printf("\nentered: %s\n", scan_string());

	fclose(fp);
	printf("Enter filename: ");
	scanf("%s", fname2);	
	if (strcmp(fname2,"stdin")!=0 && (fp=freopen(fname2, "r", stdin)) != 0)
		printf("file error\n");

		printf("Enter integer: ");
		printf("\nentered: %d\n", scan_int(0,22));
		printf("Enter double: ");
		printf("\nentered: %.6f\n", scan_double(0.0,1.0));
		printf("Enter string: ");
		printf("\nentered: %s\n", scan_string());

	fclose(fp);
	printf("Enter filename: ");
	scanf("%s", fname1);	
	if (strcmp(fname1,"stdin")!=0 && (fp=freopen(fname1, "r", stdin)) != 0)
		printf("file error\n");


		printf("Enter integer: ");
		printf("\nentered: %d\n", scan_int(0,22));
		printf("Enter double: ");
		printf("\nentered: %.6f\n", scan_double(0.0,1.0));
		printf("Enter string: ");
		printf("\nentered: %s\n", scan_string());
}
*/
