/***************************************************************************/
/****************************** NONGREEDY_STATS.C **************************/
/***************************************************************************/
#include <ctype.h>
#include <stdio.h>
#include <math.h>
#include "part.h"
#include "backtrack.h"

void free_vlist(VLISTPTR);
void bfs(int, int *);
void print_distances(int, int *);
void update_data_sets();
void update_clusters(int *);
int adj(int, int);
void print_slist(VLISTPTR v_list);
void update_dists_k_over_total(int);
void update_dists_k(int);
void update_dists_abs(int);
void update_dists_ng_10(int);
void init_stabilizer_lists();
void post_proc (int, int, int, int, int);
void print_ng_vertices();


extern int vertices[];
extern int vertex_loc[];
extern int deg[];
extern int num;
extern int sort_type ;
extern int look_ahead;
extern int output_short, output_detail, output_adjacent ;
extern struct adj_node *adj_list[];

extern int 
    hold_ng_cost[2][MAX_VERTS],     /* cost in ng upneighbors          */
    hold_node_cost[2][MAX_VERTICES],   /* keep track of best partitioning */
    hold_future_cost[2][MAX_VERTS], /* information.                    */
    hold_vertices[MAX_VERTICES],
    hold_post_tots[MAX_VERTICES],                                             
    hold_post_tots_down[MAX_VERTS],	/* 2/21/93 */
    hold_post_tots_both[MAX_VERTS], 	/* 2/21/93 */
    part_neigh_diffs[MAX_VERTS],	/* 2/21/93 */
    part_diffs[MAX_VERTS],		/* 2/21/93 */
    hold_vertex_loc[MAX_VERTICES],
    hold_even_b[MAX_VERTS],
    hold_tie_1[MAX_VERTS],  		/* conc. reord.# of ties |a-b|  */
    hold_tie_2[MAX_VERTS],  		/* conc. reord.# of tie degrees */
    hold_nongreedy,
    hold_greedy;

extern int tbestcost;			/* best of all KL iterations	*/
extern double prob;
extern FILE *outfile, *pfile ;

int Part_Match_Point;		   /* 2/21/93  The rank of the place where */
				   /* the unplaced vertices, if all placed */					   /* at the same time (greedily), go to   */
				   /* the exact KL partitions.		   */


int P_Diff;			   /* partition difference (# vert)  	*/
int strongly_ng;		   /* the number of such v  		*/
int greedy, nongreedy;		   /* total greedy/nongreedy steps   	*/
int Total_Changes;		   /* num. of v's that change p_n/p_g   */
int Non_Adj_Changes;		   /* as above, and not adj to z        */
int Last_Meaningful_Nongreedy;     /* switching its placement to     	*/
				   /* greedy worsens # cuts          	*/
				   /* Note: this is the RANK in order	*/
				   /* of placement of the last mean..	*/
int Catch_Up = EOF;
int Surpass = EOF;
int N_Prob;			/* NG_Up_Results is discrete, divided      */
				/* into rows of different edge prob. find  */
				/* out which of these is closest to prob   */

int NG_Index_Array_Count;	/* #elems in NG_Index_Array 		   */
int *NG_Index_Array;		/* keep track of sequences of g/ng steps   */
				/* and how many times they occur.  The     */
				/* constant MAX_NG_STRING indicates the    */
				/* length of the sequence to be tracked, so*/
				/* the array will have size 2^MAX_NG_STRING*/

NG_DIST Nongreedy_Distribution[MAX_VERTS];  /* ND[i] = #graphs.w/i sng  */

int Changes[MAX_DIST];		   /* at bfs level i, Changes[i] is the      */
				   /* number of vertices that change p_n/p_g */
int Neighbors_At_Level[MAX_DIST];  /* # neighbors of z at bfs level i        */

double Change_Ave, No_Change_Ave;


VLIST nongreedy_vertices[MAX_VERTS]; /* list of all nongreedy verts    */

/*** hold fields for summary information ***/


extern PCT_NG_UP (*NG_Up_Results)[NG_UP_SIZE];	/* gives ng up percentages  */
						/* from ng_uptab.c          */

int Num_NG_In_First_10;		/* # ng in first 10 steps (for use with)    */
				/* -DT and -PT options 			    */

int in_interval(double pct, double start, double step, int k)
{
	if (step >= 0.0) {
		if (pct <= (start + ((k+1)*step)))
			return(TRUE);
	} else if (step < 0.0) {
		if (pct >= (start + ((k+1)*step)))
			return(TRUE);
	}
	return(FALSE);
}

/**** 2/21/93  change post_tots to hold rank instead of vert num ****/

/******************************************************/
/************   print non_greedy vertices  ************/
/******************************************************/
int print_nongreedy()
{
	extern int NG_10_Intervals;
	extern int Abs_Diff_Intervals;
	extern int Up_NG_Intervals;
	extern int Output_Sng_Distfile;
	extern int Output_Configuration_Table;
	extern int Output_Stabilizers;
	extern int P_N_Prime;
	extern int Done_With_Alg;
	extern int Change_Index;
	extern int strongly_ng;			/* nongreedy, not even_b    */ 

	int i, j, t, highgreedy;
	int even_b_count = 0;			/* # verts. that      */
						/* had even diff, yet */
						/* went to SET_B      */
	int adj_flag;				/* TRUE if some two   */
						/* consequtive ng.    */
						/* are adjacent	      */

	int ng_index=0;			/* int to represent sequence of  */
					/* ng/g steps: e.g. 4 --> 100 for*/
					/* 1st 3 steps.			 */
   	int plot[101];
   	struct adj_node  *temp;

	if(!output_short) {
		fprintf(outfile,"\n");
		fprintf(outfile,"non-greedy vertics (num by rank)-- ");
	}
   	if(pfile != NULL)
      	    fprintf(pfile,"amount each vertex is g/ng (vert. num by rank)\n");
   	for(i=0;i<101;i++)  /* 0-49 ng, 50 even, 51-100 g */
      		plot[i] = 0 ;
   	highgreedy = VERY_LOW ;
   	for(i=0;i<num;i++) {
      		/*i = vertices[j];*/
      		if (hold_post_tots[i] < 0) { /** non-greedy vertex **/
         		if(!output_short)
	    			fprintf(outfile," %d:%d",
					vertices[i],-hold_post_tots[i]);
         		if(-hold_post_tots[i] > highgreedy)
            			highgreedy = -hold_post_tots[i];
       		}
      		if(pfile != NULL) {
	 		fprintf(pfile,"%d\n",hold_post_tots[i]);
	 		t = hold_post_tots[i] +50 ;
	 		if(t <0 ) t = 0;
	 		if(t>100) t = 100 ;
	 		plot[t]++;
       		}
		if (i < MAX_NG_STRING)
			ng_index = 2 * ng_index + (hold_post_tots[i] < 0);
    	}
	if (Output_Configuration_Table)
		if (NG_Index_Array[ng_index]++ == 0)
			NG_Index_Array_Count++;
	strongly_ng = 0;
	adj_flag = FALSE;
	Num_NG_In_First_10 = 0;
        for(j=0;j<num;j++) {
	  if (hold_post_tots[j] < 0) {
	      if (j<10)	Num_NG_In_First_10++;
	      nongreedy_vertices[strongly_ng].v = hold_vertices[j];
	      nongreedy_vertices[strongly_ng++].rank = j;
	      /********************************************/
	      /*** test if consequtive sng are adjacent ***/
	      /********************************************/
	      if ((strongly_ng >=2) && 
	          (nongreedy_vertices[strongly_ng-2].rank == j-1)) {
	            for(temp=adj_list[strongly_ng-2];temp;temp=temp->next)
	      		if (temp->node_no==nongreedy_vertices[strongly_ng-1].v)
				adj_flag = TRUE;
	      	    }		
	 }

	 if (Output_Sng_Distfile) {
	 	if (Abs_Diff_Intervals) {
			update_dists_abs(j);
	 	} else if (Up_NG_Intervals) {
			update_dists_k(j);
		} else if (NG_10_Intervals) {
			update_dists_ng_10(j);
		} else
			update_dists_k_over_total(j);
	 }
	}
   	if(!output_short) {
      		/** print table of non-greedy steps of each degree **/
      		fprintf(outfile,"\n\nNon-Greedy Degree : Vertex Number");
      		/** only print even which went to set B **/
         	fprintf(outfile,"\nEVEN B:");
         	for(j=0;j<num;j++)
            		if(hold_even_b[j]) {
               			fprintf(outfile," %4d" ,j);
				even_b_count++;
			}
      		for(i=1;i<=highgreedy;i++) {
         		fprintf(outfile,"\n %4d:",i);
         		for(j=0;j<num;j++) 
            			if(-hold_post_tots[j] == i) 
               				fprintf(outfile," %4d",vertices[j]);
       		}
      		fprintf(outfile,"\n");

		if ((strongly_ng+even_b_count) != hold_nongreedy)
			printf("Error in nongreedy count\n"), exit(32);

		/************************************************************/
		/****** the following call is ONLY to find stab. lists ******/
		/************************************************************/
		if (Output_Stabilizers)  {

		  init_stabilizer_lists();
		  if ((Change_Index < 0) || (Change_Index > num-1)) {
			for (i=0; i<num; i++)	
				vertex_loc[i] = hold_vertex_loc[i];
			post_proc(FALSE, num+1, P_N_Prime,0, 0);
			Done_With_Alg = FALSE;		/* kludge to avoid */
							/* wasted time in  */
		  }					/* get_post_cost() */

		}
		/************************************************************/

    	}
	print_ng_vertices();
	fprintf(outfile,"*************************************************\n");
	if (adj_flag)
		fprintf(outfile,"Adjacent consequtive nongreedy exist\n");
	else
		fprintf(outfile,"No consequtive nongreedy are adjacent\n");
	fprintf(outfile,"*************************************************\n");

   	if(output_detail) {
		/**** way outdated **************************/
      		/** print detail of each vertex processed ***/
      	/*	fprintf(outfile,"\nVERTEX DEGREE K-L setA setB");
      		if(sort_type == 7 || sort_type == 8)
         		fprintf(outfile," tie1 tie2");
      		if(look_ahead)
         		fprintf(outfile," futA futB");
      		if(output_adjacent)
         		fprintf(outfile,"  ADJ");
      		fprintf(outfile,"\n");
      		for(i=0;i<num;i++) {
         		if(hold_post_tots[i] < 0 || hold_even_b[i]) 
            			fprintf(outfile,"***");
         		else
            			fprintf(outfile,"   ");
         		t = hold_vertices[i];
         		fprintf(outfile,"%4d%5d  ",i,deg[t]);
         		if(hold_vertex_loc[t] == SET_A)
            			fprintf(outfile,"  A");
         		else
            			fprintf(outfile,"  B");
         		fprintf(outfile,"%5d%5d",
            		hold_node_cost[SET_B][t],
            		hold_node_cost[SET_A][t]);
         		if(sort_type == 7 || sort_type == 8)
            			fprintf(outfile,"%5d%5d",
               				hold_tie_1[i], hold_tie_2[i]);
         		if(look_ahead)
            			fprintf(outfile,"%5d%5d",
               				hold_future_cost[SET_B][t],
               				hold_future_cost[SET_A][t]);

         		if(output_adjacent) {
            			fprintf(outfile," ");
            			p = adj_list[t];
            			n = 0 ;
            			if(sort_type != 7 && sort_type != 8)
               				n = n - 2;
            			if(!look_ahead)
               				n = n - 2;
            			while(p != NULL) {
               				if(++n >= 9) {
                  			   fprintf(outfile,"\n                            ");
					   n = 1 ;
                  			   if(sort_type == 7 || sort_type == 8){
                     				fprintf(outfile,"          ");
                     				n = n-2;
		   			   }
                  			   if(look_ahead) {
                     				fprintf(outfile,"          ");
                     				n = n -2;
		   			   }
					}
               				j = 0 ;
               				while(hold_vertices[j] != p->node_no) 
                  				j++;
               				fprintf(outfile,"%4d",j);
               				p = p->next ;
	     			}
	  		}
         		fprintf(outfile,"\n");
       		}
      		fprintf(outfile,"\n\n");
    	****/
	}

	if(pfile != NULL) {
		fprintf(pfile,"XXXXXXXXXXXXXXXX\n\n");
		fprintf(pfile,"g/ng,# of vert.\n");
		for(i =0;i<101;i++)
			fprintf(pfile,"%d,%d\n",i-50,plot[i]);
	}
	return(strongly_ng);
}


void print_ng_vertices()
{
	extern int Output_Stabilizers;
	extern int strongly_ng;			/* the # ng - # even_b */
	int i, temp_rank;
	int first_100=0, first_250=0, first_400=0;/* number of strongly ng */
						  /* in first n placements */
	fprintf(outfile,"*************************************************\n");
	fprintf(outfile,"*** STRONGLY NONGREEDY VERTICES *****************\n");
	fprintf(outfile,"*************************************************\n");
	fprintf(outfile,"** Rank ** Vertex Number ************************\n");
	fprintf(outfile,"*************************************************\n");
	
	for (i=0; i<strongly_ng; i++) {
		fprintf(outfile,"%7d", nongreedy_vertices[i].rank+1);
		fprintf(outfile,"%10d  \t", nongreedy_vertices[i].v);
		if (Output_Stabilizers) {
			print_slist(nongreedy_vertices[i].next);
			/*** NEED  free_slist() ***/
		} else    fprintf(outfile,"\n");
		
		temp_rank = nongreedy_vertices[i].rank+1;
		if (temp_rank < 100)
			first_100++;
		if (temp_rank < 250)
			first_250++;
		if (temp_rank < 400)
			first_400++;
	}
	fprintf(outfile,"*************************************************\n");
	fprintf(outfile,"Total\t<400\t<250\t <100\n");
	fprintf(outfile,"-----\t----\t----\t----\n");
	fprintf(outfile,"%5d\t%4d\t%4d\t%4d\n", strongly_ng, first_400,
						first_250, first_100);
	fprintf(outfile,"*************************************************\n");
}

void init_stabilizer_lists()
/*************************************************************************/
/** init_stabilizer_lists()Each entry in the array nongreedy_vertices has*/
/**			an associated list of greedy vertices that help **/
/**			stabilize it.  In other words, when the nongr.  **/
/**			is placed, it is unstable.  At some later point **/
/**			it probably becomes stable or else it wouldn't  **/
/**			have been placed nongreedily (the exception is  **/
/**			if it was placed solely for balance).  We are   **/
/**			interested in seeing if one greedy step is ever **/
/**			a stabilizing factor for more than one nongreedy.*/
/**			The greedy lists are built with calls from      **/
/**			get_post_cost().				**/
/*************************************************************************/
{
	extern int strongly_ng;
	int i;

	for (i=0; i<strongly_ng; i++) {
		nongreedy_vertices[i].next = (VLISTPTR) malloc(sizeof(VLIST));
		nongreedy_vertices[i].next->next = NULL;
	}
}

int search_nongreedy_list(int v_num)
{
	extern int strongly_ng;
	extern VLIST nongreedy_vertices[];
	int i;

	for (i=0; i<strongly_ng; i++)
		if (nongreedy_vertices[i].v == v_num)
			return(i);
	return(EOF);
}

VLISTPTR new_vlist()
{
	VLISTPTR vlist = (VLISTPTR) malloc(sizeof(VLIST));
	vlist->next = NULL;
	return vlist;
}
	
void insert_vlist(int v_num, int v_rank, int v_diff_p_n, int v_diff_p_g, 
		  VLISTPTR v_list)
{
	VLISTPTR vtemp = (VLISTPTR) malloc(sizeof(VLIST));
	
	vtemp->v = v_num;
	vtemp->rank = v_rank;
	vtemp->diff_p_n = v_diff_p_n;
	vtemp->diff_p_g = v_diff_p_g;
	vtemp->next = v_list->next;
	v_list->next = vtemp;
}

void free_vlist(VLISTPTR v_list)
{
	VLISTPTR temp, temp1;
	
	if (!v_list) return;
	temp1 = v_list->next;
	while (temp1 != NULL) {
		temp = temp1;
		temp1 = temp1->next;
		free((char *) temp);
	}
	free((char *) v_list);
}

void print_slist(VLISTPTR v_list)
{
	VLISTPTR temp;
	
	if (v_list == NULL) {
		fprintf(outfile, "\n"); return;
	}
	temp = v_list->next;
	while (temp != NULL) {
		if (temp->v) 	{		/* used as a flag to indicate*/
		                                /* the vertex that brings the*/
		                  	 	/* ng's cost back to 0       */
		fprintf(outfile,"%d* ",temp->rank+1);
		temp = temp->next;		

		} else {
			fprintf(outfile,"%d ",temp->rank+1);
			temp = temp->next;		
		}
	}
	/**
	for (temp = v_list->next; temp!=NULL; temp = temp->next); 
		fprintf(outfile,"%d ",temp->rank+1);
	**/
	fprintf(outfile, "\n-----------------------------------------\n");
}

void print_change_lists(VLISTPTR p_n_list, VLISTPTR p_g_list)
{
	extern int P_N_Prime;    /* true if comparing P_n(z) and P_n'(z) */
	char     p_str[8];
	VLISTPTR temp_p_n, temp_p_g;
	
	if (P_N_Prime)		strcpy(p_str, "P_n'(z)");
	else			strcpy(p_str, "P_g(z)");

	fprintf(outfile, "-----------------------------------------\n");
	fprintf(outfile, "Differences\n");
	fprintf(outfile, "-----------------------------------------\n");
	fprintf(outfile, "P_n(z)\t\t\t | \t %s \n", p_str);
	fprintf(outfile, "-----------------------------------------\n");
	for (temp_p_n = p_n_list->next, temp_p_g = p_g_list->next; 
	     temp_p_n && temp_p_g; 
	     temp_p_n = temp_p_n->next, temp_p_g = temp_p_g->next) {
		fprintf(outfile, "%d, (rank=%d)",temp_p_n->v, temp_p_n->rank+1);
		fprintf(outfile, ", (diff=%d)",temp_p_n->diff_p_n);
		fprintf(outfile,"\t\t%d, (rank=%d)",
						 temp_p_g->v,temp_p_g->rank+1);
		fprintf(outfile, ", (diff=%d)",temp_p_g->diff_p_g);
		fprintf(outfile,"\n");
	}
	fprintf(outfile, "-----------------------------------------\n");
}
		
int comp_partitions(int *part1, int *part2)
{
	int i, diff, same;

	diff = num;
        for (i=num-1; i>=0; i--) {
                same = (part1[i] == part2[i]);
                diff -= same;
	}
        if (num-diff < diff)
                diff = num-diff;
        return(diff);
}
		
VLISTPTR compare_partitions(int *part1, int *part2, int *verts, int *p_diff)
{
	extern int abs_diff_p_n[];		/* abs_diff in p_n(z) place.*/
	extern int abs_diff_p_g[];		/* abs_diff in p_g(z) place.*/
	int i, diff, same;
	VLISTPTR v_list = (VLISTPTR) malloc(sizeof(VLIST));

	v_list->next = NULL;

	diff = num;
        for (i=num-1; i>=0; i--) {
                same = (part1[verts[i]] == part2[verts[i]]);
                diff -= same;
		if (!same) 
			insert_vlist(verts[i], i, abs_diff_p_n[verts[i]], 
				abs_diff_p_g[verts[i]], v_list);
	}
        if (num-diff < diff)
                diff = num-diff;
	*p_diff = diff;
        return(v_list);
}

void compare_situations(int *p_n_order, int *p_n_locs, int *p_g_order, 
			int *p_g_locs, int lmng)
{
	extern int distance[]; 	/* computed in bfs() */
	int i,j; 
	int current_p_n_vertex;
	int current_p_g_vertex;
	int a_n_count, b_n_count, a_g_count, b_g_count;
	int a_n_match, b_n_match, a_g_match, b_g_match;
	int a_intersection=0, b_intersection=0;
	int size_of_union;
	int no_changes=0;
	int a_n[MAX_VERTS];    /* set of verts. in SET_A at some time */
	int b_n[MAX_VERTS];    /* set of verts. in SET_B at some time */
	int a_g[MAX_VERTS];    /* same for partition P_g(z)           */
	int b_g[MAX_VERTS];
	int change_total=0.0, no_change_total=0.0;
	double intersection_ratio[MAX_VERTS];

	Total_Changes=0;
	for (i=0; i<num; i++) {
		a_n[i] = 0;
		b_n[i] = 0;
	}

	for (i=lmng; i<num-1; i++) {
	       	current_p_n_vertex = p_n_order[i];
		if (p_n_locs[current_p_n_vertex] == SET_A)
			a_n[current_p_n_vertex] = 1;
		else
			b_n[current_p_n_vertex] = 1;
		current_p_n_vertex = p_n_order[i+1];

		/*******************************************************/
		/*** now we know the situation of current_p_n_vertex ***/
		/*** in p_n(z).					     ***/
		/*** this vertex is in a different placement position***/
		/*** in p_g(z).  Therefore, we need to find this     ***/
		/*** vertex in p_g(z) and then we'll know it's 	     ***/
		/*** situation there.				     ***/
		/*******************************************************/

		for (j=0; j<num; j++)
			a_g[j] = b_g[j] = 0;

		for (j=lmng; p_g_order[j] != current_p_n_vertex; j++) {
	       		current_p_g_vertex = p_g_order[j];
			if (p_g_locs[current_p_g_vertex] == SET_A)
				a_g[current_p_g_vertex] = 1;
			else
				b_g[current_p_g_vertex] = 1;
		}

		a_n_count = b_n_count = a_g_count = b_g_count = 0;
		a_intersection = b_intersection = 0;
		for (j=0; j<num; j++) {
			a_n_match = b_n_match = a_g_match = b_g_match = FALSE;
			if ((a_n[j] == 1) && adj(current_p_n_vertex, j)) {
				a_n_count++;
				a_n_match = TRUE;
			}  
			if ((a_g[j] == 1) && adj(current_p_n_vertex, j)) {
				a_g_count++;
				a_g_match = TRUE;
			}  
			if ((b_n[j] == 1) && adj(current_p_n_vertex, j)) {
				b_n_count++;
				b_n_match = TRUE;
			} 
			if ((b_g[j] == 1) && adj(current_p_n_vertex, j)) {
				b_g_count++;
				b_g_match = TRUE;
			} 
			if (a_n_match && a_g_match)  
				a_intersection++;
			if (b_n_match && b_g_match)  
				b_intersection++;
		}
		size_of_union = MAX(a_n_count, a_g_count) + 
				MAX(b_n_count, b_g_count);
		if (size_of_union == 0)
			intersection_ratio[current_p_n_vertex] = 0;
		else
			intersection_ratio[current_p_n_vertex] = 
			  (double) (100 * (a_intersection + b_intersection)) / 
			  (double) size_of_union;

	}
	for (i=0; i<MAX_DIST; i++) {
		Changes[i] = 0;
		Neighbors_At_Level[i] = 0;
	}
	/******** find averages *******/
	for (i=lmng+1; i<num; i++) {
		j = p_n_order[i];
		if (p_n_locs[j] == p_g_locs[j]) {
			no_change_total += intersection_ratio[j];
			no_changes++;
		} else {
			change_total += intersection_ratio[j];
			Total_Changes++;
			Changes[distance[j]]++;
		}
		Neighbors_At_Level[distance[j]]++;
		
	}
	if (lmng < num) {
	  if (Total_Changes > 0) {

	  Change_Ave = change_total / (double) Total_Changes;
	  fprintf(outfile,"ave %% of intersection, changed vertices: %.6f%%\n",
			Change_Ave);
	  } else {

	  Change_Ave = 100.0;
	  fprintf(outfile,"ave %% of intersection, changed vertices: 100%%.\n");

	  }

	  if (no_changes > 0.0) {

	  No_Change_Ave = no_change_total / (double) no_changes; 
	  fprintf(outfile,"ave %% of intersection, unchanged vertices: %.6f%%\n",
			No_Change_Ave);

	  } else {

	  No_Change_Ave = 0.0;
	  fprintf(outfile,"ave %% of intersection, unchanged vertices: N.A.\n");
	  
	  }
	}
	Total_Changes += 1;    /* add_to_summary() counts the lmng 	     */
}

void print_ng_vectors()
/*** 2/21/93 ***/
{
	extern int Pretend_Changes[];
	extern int Part_Match_Point;

	int i, ng1=0, ng2=0, ng3=0;
	int flag=FALSE;
	FILE *fp; 

	Part_Match_Point=num;
	
	if ((fp = fopen("ng_vects2", "w")) == NULL)
		printf("Error: can't open ng_vects\n"), exit(1);
	
	for (i=0; i<num; i++) {
		fprintf(fp, "%7d%7d%7d%7d%7d%7d", i, hold_post_tots[i], 
				hold_post_tots_down[i], hold_post_tots_both[i],
				part_neigh_diffs[i], part_diffs[i]);
		if (!flag && (part_diffs[i] == 0)) {
			fprintf(fp,"***");
			flag = TRUE;
			Part_Match_Point = i;
		}
		fprintf(fp,"%7d\n", Pretend_Changes[hold_vertices[i]]);
		if (hold_post_tots[i] < 0) ng1++;
		if (hold_post_tots_down[i] < 0) ng2++;
		if (hold_post_tots_both[i] < 0) ng3++;
	}
	fprintf(fp, "------------------------------------------------------\n");
	fprintf(fp, "%14d%7d%7d%14d\n", ng1, ng2, ng3, Part_Match_Point);
	fclose(fp);
}

void find_mutual_neighbors()
{
	extern int hold_vertex_loc[];
	int i, j, mut_neighs;
	int same_count=0, diff_count=0;
	int same_tots=0, diff_tots=0;
	struct adj_node *temp;
	
	fprintf(outfile, "*******************************************\n");
	fprintf(outfile, "* Mutual Neighbors ************************\n");
	fprintf(outfile, "*******************************************\n");
	for (i=0;i<num;i++)
		for (j=0;j<num;j++) {
			if (i==j) continue;
			mut_neighs = 0;
			for (temp=adj_list[i]; temp; temp=temp->next)
				if (adj(j, temp->node_no))
					mut_neighs++;
			if (hold_vertex_loc[i] == hold_vertex_loc[j]) {
				same_tots += mut_neighs;
				same_count++;
			} else {
				diff_tots += mut_neighs;
				diff_count++;
			}
		}
	fprintf(outfile,"same partition ave: %.6f\n",(double)same_tots/same_count);
	fprintf(outfile,"diff partition ave: %.6f\n",(double)diff_tots/diff_count);
	fprintf(outfile, "*******************************************\n");
}

void nongreedy_stats()
{
	extern int Find_Clusters;     /* see cluster.c                 */
	extern int Compute_Variance; /* see data_sets.c                */
	extern int Abs_Distfile;     /* TRUE => graph of abs_diffs     */
	extern int P_N_Prime;        /* TRUE if comparing P_n vs. P_n' */
	extern int p_graph;          /* produce plotfiles if true      */
	extern int strongly_ng;
	extern int Tot_Cost[];
	extern int hold_vertices[];
	extern int abs_diff_p_n[];
	/* extern int abs_diff_p_g[];*/

	int i,j; /*k, temp, p_g_rank; */
	int flag;
	/*int p_n_increase;*/		   /* how many cuts are adding when  */
					   /* the current vertex is placed(pn*/
	/*int p_g_increase;*/		   /* how many cuts are adding when  */
					   /* the current vertex is placed(pg*/
	/*int gr_benefitting_from_lmng=0;*//* how many greedy verts benef?   */
	/* int gr_hurt_by_lmng=0;*/	   /* how many suffer losses	     */
	/*int gr_amount_of_benefit=0;*/    /* what is the total benefit?     */
	/*int gr_amount_of_loss=0;*/	   /* what is the total loss?        */
	char p_str[8];			   /* "P_g(z)" or "P_n'(z)"          */
	int p_n[MAX_VERTS]; 		   /* array of cum. cuts in P_n(z)   */
	/*int p_g_ranks[MAX_VERTS];*/	   /* rank of i in 'vertices' (P_g)  */
	VLISTPTR p_n_changed, p_g_changed; /* p_g_changed does double duty   */
					   /* holds p_n' changed also        */
	FILE *p_n_plotfile, *other_plotfile, *abs_distfile;

	/*find_mutual_neighbors();*/

	if (Find_Clusters)
		update_clusters(hold_post_tots);

	if (Compute_Variance)
		update_data_sets();

	if (Abs_Distfile) {
		if ((abs_distfile = fopen("abs.dist", "w")) == NULL)
			printf("Error opening abs.dist\n"),exit(23);
		for (i=0; i<num; i++)
			fprintf(abs_distfile,"%d %d\n", i, 
				abs_diff_p_n[vertices[i]]);
		fclose(abs_distfile);
	}
	if (P_N_Prime) 	strcpy(p_str,"P_n'(z)");
	else  		strcpy(p_str,"P_g(z)");

	Last_Meaningful_Nongreedy = num;
	strongly_ng = print_nongreedy();
	for (i=0; i<num; i++)	{			/* restore best part.*/
		vertex_loc[i] = hold_vertex_loc[i];
		p_n[i] = Tot_Cost[i];
	}

	for (i=strongly_ng-1; i>=0; i--) { 
	/*for (i=0; i<=strongly_ng-1; i++) { ******************/
	   post_proc(FALSE, nongreedy_vertices[i].rank, P_N_Prime, 0, 0);
	   if (Tot_Cost[num-1] < tbestcost)  {
		Last_Meaningful_Nongreedy = nongreedy_vertices[i].rank;
		fprintf(outfile,"LAST MEANINGFUL NONGREEDY:%dth (vertex %d)\n",
				Last_Meaningful_Nongreedy+1, 
				nongreedy_vertices[i].v);
		p_n_changed    = compare_partitions(vertex_loc, 
						    hold_vertex_loc, 
						    hold_vertices, &P_Diff);
		p_g_changed    = compare_partitions(vertex_loc, 
						    hold_vertex_loc, 
						    vertices, &P_Diff);
		if (!output_short)
		     print_change_lists(p_n_changed, p_g_changed);
		bfs(Last_Meaningful_Nongreedy, hold_vertices);
		if (!output_short)
		     print_distances(Last_Meaningful_Nongreedy, hold_vertices);
		fprintf(outfile,"Partition Difference P_n(z)/%s: %d\n",
						p_str, P_Diff);
		free_vlist(p_n_changed);
		free_vlist(p_g_changed);
		break;
	   } /* else  if (Tot_Cost[num-1] == tbestcost) {          */
		/**********************************************************/
		/*** We have a new partitioning with the same cost, but ***/
		/*** 1 fewer strongly nongreedy vertex after z.  We'll  ***/
		/*** call this new partition our "best" kl partition.   ***/
		/*** After this loop, no strongly ng will occur after z.***/
		/**********************************************************/
		/*for (j=0; j<num; j++)	{
			hold_vertex_loc[j] = vertex_loc[j];
			p_n[j] = Tot_Cost[j];
		}
	   } else
		fprintf(outfile, "OOOOPS\n");*/
	   for (j=0; j<num; j++)
		vertex_loc[j] = hold_vertex_loc[j];
	}

	/**********************************************************/
	/** Test each greedy after z (the lmng) to see if things **/
	/** would have been better if it had been placed in a    **/
	/** nongreedy way.                                       **/
	/**********************************************************/
	flag = FALSE;
	/*
	for (i=Last_Meaningful_Nongreedy+1; i<num; i++) {
		for (j=0; j<num; j++)
			vertex_loc[j] = hold_vertex_loc[j];
		post_proc(FALSE, i, P_N_Prime);
		if (p_n[num-1] < Tot_Cost[num-1])
			flag = TRUE;*/		  /* some greedy would be */
						  /* better off nongreedy */
		/***was already commented out*****for (k=0; k<num; k++)  
			printf(" %d\t%d\t%d\n", vertices[k]+1,
				 Tot_Cost[k], vertex_loc[k]);
		getchar();
		*/
	/*}*/					  /* lead to better part. */
						  /* if it had gone ng.   */
	fprintf(outfile, "---------------------------------------------- \n");
	fprintf(outfile, "Greedy Vertices after z  \n");
	fprintf(outfile, "---------------------------------------------- \n");
	if (flag)
		fprintf(outfile, "Some greedy would be better nongreedy\n");
	else
		fprintf(outfile, "All greedy after z are ok\n");
	fprintf(outfile, "---------------------------------------------- \n");

		/*************************************************/
	        /* this effectively makes P_g into P_n and vice **/
	  	/* versa.  The first greedy after the lmng is   **/
		/* changed to nongreedy, and the benefits are   **/
		/* examined.  NOTE:  Now the "benefits are the  **/
		/* costs and vice versa.  (because P_g is now   **/
		/* the "old" partitioning and P_n is the "new"  **/
		/* partitioning.)  REMEMBER TO REMOVE THIS!!!!  **/
	/*********************************************************/

	/*****************************************************************
	fprintf(outfile, "---------------------------------------------- \n");
	fprintf(outfile, "Greedy Vertices benefitting from n.g. place. : \n");
	fprintf(outfile, "of 1st vertex after l.m.n.g                  : \n");
	fprintf(outfile, "---------------------------------------------- \n");
	 
	for (i=0; i<num; i++) p_g_ranks[vertices[i]] = i; 
	for (i=Last_Meaningful_Nongreedy+1; i<num; i++) {
		temp = hold_vertices[i];
		p_g_rank = p_g_ranks[temp];
		p_n_increase = (p_n[i]-p_n[i-1]);
		p_g_increase = Tot_Cost[p_g_rank] - Tot_Cost[p_g_rank-1];
		if (p_n_increase > p_g_increase) {
			gr_benefitting_from_lmng++;
			gr_amount_of_benefit += p_n_increase - p_g_increase;
			if (!output_short)
				fprintf(outfile, " %d", temp);
		} else if (p_n_increase < p_g_increase) {
			gr_hurt_by_lmng++;
			gr_amount_of_loss += p_g_increase - p_n_increase;
		}
	}
	fprintf(outfile, "\n---------------------------------------------- \n");
	fprintf(outfile, "TOTAL *HURT:        %d, total *loss: %d\n", 
			gr_benefitting_from_lmng, gr_amount_of_benefit);
	fprintf(outfile, "TOTAL *BENEFITTING: %d, total benefit: %d\n", 
			gr_hurt_by_lmng, gr_amount_of_loss);
	fprintf(outfile, "---------------------------------------------- \n");
	*/	
	/*****GP DEBUGGING ONLY: *******/
	/*for (i=0; i<num; i++)  printf(" %d\t%d\t%d\n", vertices[i]+1,
				 Tot_Cost[i], vertex_loc[i]);*/
	/*****GP DEBUGGING ONLY: *******/

	/*****************************************************************/

	/*****************************************************************/
	/*fprintf(outfile, "---------------------------------------------- \n");
	fprintf(outfile, "TOTAL COST (P_n): %d\n", p_n[num-1]);
	fprintf(outfile, "TOTAL COST (P_g) (lmng+1 nongreedy): %d\n", 
					Tot_Cost[num-1]);
	fprintf(outfile, "---------------------------------------------- \n");
	*/
	/******************************************************************/
	/*compare_situations(hold_vertices,hold_vertex_loc,vertices,vertex_loc,
				Last_Meaningful_Nongreedy);*/
	/******************************************************************/
	for (i=Last_Meaningful_Nongreedy; i<num; i++) {
		if (Tot_Cost[i] <= p_n[i]) {
			Catch_Up = i;
			break;
		} else if (i == num-1)
			Catch_Up = num;
	}
	if (Catch_Up > EOF)
		fprintf(outfile, "catch up point: %d\n", Catch_Up+1);

	for (i=Catch_Up; i<num; i++)
		if (Tot_Cost[i] < p_n[i]) {
			Surpass = i;
			break;
		} else if (i == num-1)
			Surpass = num;
	if (Surpass > EOF)
		fprintf(outfile, "Surpass point: %d\n", Surpass+1);

	/*for (i=Surpass; i<num; i++)
		if (Tot_Cost[i] > p_n[i]) 
			fprintf(outfile, "p_g pulls ahead again: %d\n", i+1);*/

	Catch_Up += 1;   /* these next 3 are C subscript values.  we */
	Surpass += 1;    /* want real ranks to count in summary.     */
	Last_Meaningful_Nongreedy += 1;

	if (p_graph) {
		p_n_plotfile = fopen("p_n_plot", "w");
		if (P_N_Prime)
			other_plotfile = fopen("p_n_prime_plot", "w");
		else
			other_plotfile = fopen("p_g_plot", "w");
		for (i=0; i<num; i++) {
			fprintf(p_n_plotfile," %d %d\n", i, p_n[i]);	
			fprintf(other_plotfile," %d %d\n", i, Tot_Cost[i]);	
		}
		fclose(p_n_plotfile);
		fclose(other_plotfile);
	}
}
