#ifndef PART_H
#define PART_H
/***********************************************************/
/*                  Graph Paritioning                      */
/*                       part.h                            */
/*                                                         */
/*                Author:  James P. Fox                    */
/*                Advisor: Mark Goldberg                   */
/*                                                         */
/*                    Masters Project                      */
/*            Rensselaer Polytechnic Institute             */
/*                  Summer, Fall, 1990                     */
/***********************************************************/
/*     This file contains the definitions and declarations */
/*  which are common to the different modules in the graph */
/*  partitioning programs.                                 */
/***********************************************************/
#include <math.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

#define DISTRIB_SIZE		100
#define NG_UP_SIZE		100
#define MAX_INT 		2147483647
#define MAX_VERTS 		100000	/* originally 2000 */
#define MAX_VERTICES 		100000  /* originally 200000 */
#define MAX_IND_SIZE 		1500
#define MAX_NG_STRING 		10.0		/** see NG_Index_Array, in **/
						/** file nongreedy_stats.c **/
#define NG_INDEX_SIZE		((int) pow(2.0, MAX_NG_STRING))
#define MAX_DEGREE_NONGREEDY 	10		
#define MAX_PASSES 		3125
#define MAX_DIST 		20
#define MAX_TIME		524288.00	/** longest heuristic run   **/
#define MAX_REG_TRY		100		/** # times to try to build **/
						/** regular graph	    **/
#define MAX_GRAPH_HEIGHT 	20
#define MAX_GRAPH_WIDTH 	50

#define CREATE_GRAPH   		1
#define DONT_CREATE_GRAPH    	0
#define ALLOC    		1
#define DONT_ALLOC	    	0
#define RESET_LOCS    		1
#define DONT_RESET_LOCS    	0
#define FOUND          		1
#define NOT_FOUND      		2
#define END_OF_LIST    		9999
#define SKIP_VERTEX    		-99
#define VERY_HIGH      		999999
#define VERY_LOW       		-999999
#define SET_A          		0
#define SET_B          		1
#define RESET_DEG		1
#define DONT_RESET_DEG		0

#define POINT05			0	/* indices into sng_dist[i][] */
#define POINT10			1	/* these are used by PG on    */
#define POINT15			2	/* geometric graphs	      */
#define POINT20			3
#define POINT25			4

#define POINT1			0	/* indices into sng_dist[i][] */
#define POINT2			1	/* these are used by PG on    */
#define POINT3			2	/* random graphs	      */
#define POINT4			3
#define POINT5			4

#define INITPOOLSIZE   		65536
#define AUGMENT        		1
#define LOCKED        		1
#define NOT_LOCKED          	0

#define PASS 			1
#define FAIL			-1

#define PRINT 			1
#define DONT_PRINT		0

#define GREEDY			0	/* reversed from 1-0  6/6/93 */
#define NONGREEDY		1

#define TRUE 			1
#define FALSE 			0

#define MODE_MAX 		-239
#define MODE_MIN 		-240
#define MIN_QUOTIENT 		-241

#define ALG_TYPE_KL 		-10
#define ALG_TYPE_NEAR_GREEDY 	-20
#define BACKTRACKING 		-30
#define RANDOM_WALK 		-40

#define SUBGRAPH_MAXDEGREE	10	/* sort type		*/

#define BACKTRACK		-50	/* sort type 		*/

#define xMULTIPLIER 		1
#define MULTIPLIER  		((int) (prob*10.0) + 1)
#define UNBALANCE		4

#define NUM_INTERVALS		0
					/* when we look at ng-up-% of verts */
					/* we see a split between the stats */
					/* for greedy and ng vertices.  We  */
					/* want to use this in our PG alg,  */
					/* so we break the range between the*/
					/* two figures into NUM_INTERVALS   */
					/* pieces			    */
#define SIZE_INTERVAL_ARRAY 	(NUM_INTERVALS+1)

#define ABS(x) (((x) >= 0) ? (x) : (-(x)))

#define MAX(x,y) ( ((x) >= (y)) ? (x) : (y))
#define MIN(x,y) ( ((x) >= (y)) ? (y) : (x))

struct adj_node {  			/** store the adjaceny list **/
   int node_no ;
   struct adj_node *next ;
};

typedef struct {			/* NONGREEDY vertex information */
	int v;				/* vertex number		*/
	int rank;			/* rank in order of placement   */
} NG_INFO;

typedef struct tlist *TLISTPTR;

typedef struct tlist {
	int 	v,			/** vertex number 		**/
		cuts,			/** from this vertex to rest of **/
					/** list			**/
		noncuts;		/** from this vertex to rest of **/
					/** list			**/
	/*TLISTPTR next;          	*/
} TLIST;

typedef struct vlist *VLISTPTR;

typedef struct vlist {
	int 	v,			/** vertex number 		**/
		rank,			/** in order of placement	**/
		diff_p_n,		/** absolute diff when placed P_n*/
		diff_p_g;		/** absolute diff when placed P_g*/
	VLISTPTR next;          
} VLIST;

typedef struct list *LISTPTR;

typedef struct list {
	int 	node_no,		/** corresponds to graph node # **/
		loc,			/** location (partition)	**/
		bucket_no,		/** adjusted cost difference	**/
		cost,			/** true cost difference	**/
		fix_next,		/** 3/20/93 At the beginning of the **/
					/** next pass, fix this node    **/
					/** and run KL on the unfixed.  **/
		lock;			/** processed?  (for kl)	**/
	LISTPTR prev, next;		/** doubly-linked list		**/
} LIST;

typedef struct pushlist *PUSHLISTPTR;

typedef struct pushlist {
	int 	side;			/** which side did it get pushed**/
	PUSHLISTPTR next;			
} PUSHLIST;

/************************************************************************/
/** The following are used in the context of a vertex placement.  For  **/
/** the i'th placement, keep the following stats as running totals over**/
/** all graphs processed so far:  <ng_up_g>:  for every time  step i   **/
/** was greedy, count the number of nongreedy upneighbors. ng_up_g is  **/
/** the total over all graphs. <up_g>: similar for ALL steps, not just **/
/** greedy steps.    <ng_up_n>, <up_n> similar, consider nongreedy steps*/
/************************************************************************/
typedef struct {
        int     times_ng,	/* for each vertex rank, how many */
		ng_up_g, 	/* times was it placed ng?        */
                up_g,   
                ng_up_n, 
                up_n;   
	double	pct_ng_up_g,
		pct_ng_up_n;
} UP_INFO, *UP_INFOPTR;

/*************************************************************************/
/** This will probably be temporary.                                    **/
/*************************************************************************/
typedef struct {
	double	gr_pct,		/* of all the times vertex i was greedy, */
				/* average percentage of ng upneighbors  */
		ng_pct;		/* of all the times vertex i ws ng,      */
				/* average percentage of ng upneighbors  */
} PCT_NG_UP, *PCT_NG_UPPTR;

/*************************************************************************/
/** This too.  It was used to CREATE the nongreedy distributions for    **/
/** various k/total values.  It is NOT used to contain the final        **/
/** <NUM_INTERVALS>							**/
/** The last entry is the "standard ng probability," the probability of **/
/** ng when all intervals are ignored & treated as one.			**/
/**  ex.     1 10  0 0   0 0   1 3  2 5   ***4 18***			**/
/*************************************************************************/
typedef struct {
	int  	num_ng[SIZE_INTERVAL_ARRAY],
		tot[SIZE_INTERVAL_ARRAY];
} NG_DIST, *NG_DISTPTR;

/*************************************************************************/
/** THIS one is actually used by PG during its execution                **/
/*************************************************************************/
typedef struct {
	double 	d[SIZE_INTERVAL_ARRAY];
} PG_DIST, *PG_DISTPTR;



int read_data() ;

void process_parameters(int, char **);

long create_graph() ;

double rrandom() ;

void print_adj_list() ;

void free_adj_list() ;

double log_time() ;

void post_processing();

void print_non_greedy();

void print_graph();

void clear_summary();

void print_summary();

void add_to_summary();

void process_reorder();

int equal_cost();

int evaluate_non_greedy();

int in_interval(double, double, double, int);

double find_prob_k_over_total(int, int*);
double find_prob_k(int, int*);
double find_prob_abs(int, int*);
double find_prob(int, int*);
double fixed_prob();

void insert_vlist(int v_num, int v_rank, int v_diff_p_n, int v_diff_p_g, 
			VLISTPTR v_list);
void init_vlist(VLISTPTR vlist);


void free_adj_mat(short **, int);
int is_eligible(int, int, int);
int is_eligible(int, int, int);
void add_adj_entry(int, int);
int calc_num_eligible (int, int, int, int);

void init_extra_file_info();
void write_sng_file(int num_v);
void write_ng_upfile(int num_v);

void alloc_buckets(int);
void free_buckets();
void init_buckets(int);
void reset_buckets(int [][MAX_VERTICES]);
void insert_to_bucket(int, int, LISTPTR);
void delete_from_bucket(LISTPTR);
void move_to_bucket(int, int, LISTPTR);
int adj(int, int);
LISTPTR bucket_extract_max(int);
LISTPTR bucket_extract_min(int);

void kl_augment(int *);
void init_local_opt(int *, int *);
void local_opt();
void bucket_move_vertex(LISTPTR);

#ifndef CC_COMPILER

/*int    printf(const char*, ...);
int    fprintf(FILE *, const char*, ...);
int    scanf(const char*, ...);
int    fscanf(FILE *, const char*, ...);
int    sscanf(char *, const char*, ...);
void   rewind(FILE *);
int    fflush(FILE *);
int    fclose(FILE *);
char   *getenv();
double tan(double);*/



#endif



#endif
