/*****************************************************************************/
/**                           regular.c					    **/
/*****************************************************************************/
#include<stdio.h>
#include "part.h"
extern short **adj_mat;
extern int deg[];
extern long lrand48();   /* 4/27/93*/

extern int num;

int create_reg_graph(int, int, int, int);

int create_unbalanced_reg_graph(int k1, int k2)
{
    while (1) {
	if (create_reg_graph(k1, 0, num/2 - 1, 0) != 1) {
		/*perror("1:");*/
		continue;
	} 
	if (create_reg_graph(k1, num/2, num - 1, 0) != 1) {
		/*perror("2:");*/
		continue;
	} 
	if (create_reg_graph(k1+k2, 0, num/2 - 1, 1) != 1) {
		/*perror("3:");*/
		continue;
	}
	break;
    }
    return(1);
}

/*****************************************************************************/
/* CREATE_REG_GRAPH							    **/
/*****************************************************************************/
/* This routine gets the needed paramater for generating a random k-regular **/
/* graph on n vertices.   It then calls routines to generate the graph.     **/
/* It returns TRUE if it was able to generate the requested graph, and FALSE**/
/* otherwise.								    **/
/* 3/26/91 - Changed to allow backing up with '!'.			    **/
/*****************************************************************************/
int create_reg_graph(int k, int low, int high, int jump)
{
	extern int fill_matrix_regular(int, int, int, int);
	extern int Unbalanced_Regular_Graphs;
	int attempts, last_attempt;
	int n = (jump) ? 2 * (high-low+1) : high-low+1;
	static int gcount;
	 
	/* try to gen the graph MAX_REG_TRY times (or until success.)*/

	if ( ((n % 2) != 0)  && ((k % 2) != 0) )
		printf("Create Graph: Error, both num and k are odd\n"),exit(1);

	last_attempt = FALSE;
	attempts = 0;  /* Number of failed attempts at generating this Graph. */
	 
	while ( attempts <= MAX_REG_TRY ) {
		if ( fill_matrix_regular(k,low,high,jump) == FALSE ) {
			printf("** Attempt failed. Will try %d more times.\n", 
				 MAX_REG_TRY-attempts);
			attempts++;
			/*perror("before free_adj_list()");*/
			free_adj_list();
			/*perror("after free_adj_list()");*/
			free_adj_mat(adj_mat, RESET_DEG);
			/*perror("after free_adj_mat() (failure)");*/
			printf("Graph: %d\n", gcount);
		} else {
			last_attempt = attempts + 1;
			if (!Unbalanced_Regular_Graphs ||
			     (Unbalanced_Regular_Graphs && (jump == 1)
						&& (last_attempt == 1))) {
				/*perror("before free_adj_mat() (success)");*/
				free_adj_mat(adj_mat, DONT_RESET_DEG);
				/*perror("after free_adj_mat() (success)");*/
				printf("Graph: %d\n", gcount);
			}
			break;
		}
	}
	gcount++;
	if (last_attempt > MAX_REG_TRY) last_attempt = FALSE;
	return(last_attempt); /* Signifies if generation was successful. */
}

/* ---------------------------------------------------------------------- */

/**************************************************************************/
/** FILL_MATRIX_REGULAR ***************************************************/
/**************************************************************************/
/**   This function fills the adjacency matrix at the first 'num'	 **/
/**   vertices with edges randomly, trying to make a k-regular graph.	 **/
/**************************************************************************/
int fill_matrix_regular (int k, int low, int high, int jump)
{
	extern int numedges;

	int ct, v, num_eligible;
	int all_ok = TRUE;
	int randnum;
	int start=low;
	int finish=high;
	int n = high-low+1;

	if (jump) {			/** make bipartite-regular **/
		start = n - 1;
		finish = finish + n;
		high += 1;
	} 

	/* For each v, connect it to k vertices of higher number. */
	for (v=low; v < high; v++) {
		while ( deg[v] < k ) {
			num_eligible=calc_num_eligible(v, k, start, finish);
			if ( num_eligible == 0 ) {
				all_ok = FALSE;
				break;
			}

			randnum = (lrand48() % num_eligible) + 1;   /* 4/27/93*/

			if (!jump)  start = v;

			for (ct=(start+1); ct <= finish; ct++) {
				if ( is_eligible(v, ct, k) ) {
					randnum = randnum - (k - deg[ct]);
					if ( randnum <= 0 ) {
						add_adj_entry(v,ct);
						add_adj_entry(ct,v);
						numedges++;
						break;
					}
				}
			}
		}

		if ( all_ok == FALSE )
		break;
	}

	/* Last v might not have proper degree. */
 	if ( deg[finish-1] != k )
    		all_ok = FALSE;

 	return(all_ok);
}
   
/******************************************************************************/
/* CALC_NUM_ELIGIBLE							     **/
/******************************************************************************/
/**   This routine calculates the sum over all the vertices that are numbered**/
/**   higher than 'vertex', but that are not adjacent to 'vertex'.	     **/
/**   This number is the number of edges still needed to make its degree == k.*/
/******************************************************************************/
int calc_num_eligible (int vertex, int k, int low, int high)
{
	int answer = FALSE;
	int ct;
	int n = high-low+1;
  
        if (vertex == 0) return(k * (n-1) - deg[vertex]);

	for ( ct = (low + 1); ct <= high; ct++ ) {
		if (adj_mat[vertex][ct] == 0) 
			answer = answer + (k - deg[ct]);
	}

	return(answer);
}

/*************************************************************************/
/** IS_ELIGIBLE								**/
/*************************************************************************/
/**   This routine returns TRUE if vertex j is elig. to be incident to  **/
/**   vertex v.								**/
/*************************************************************************/
int is_eligible (int v, int j, int k)
{
	if ( ( deg[j] < k ) && ((v == 0) || ( adj_mat[v][j] == FALSE )) )
		return(TRUE);
	else
		return(FALSE);
}
