#include <stdio.h>
#include <math.h>
#include "part.h"

#define MIN_TEMP  .00001
#define MAX_TEMP  100.0
#define QUOT_MULT 100.0 
#define COMPUTE   -3948
#define LOOKUP    -3949

void save_and_postproc(int, int, int, int *, int *);
void print_iteration(int, int, int, int, int);
void balance_partitioning();
int check_cost(int *, int);
void randinit(unsigned);
void line_randomize();
void randomize_vertices();
int compute_costs();
int one_walk(int, int, double *);
void init_walk(int);
double exp(double x);
double rrandom();
long random();

extern double time_;
extern int num;
extern int currentcost, currentminsize;
extern int tbestcost, tbestminsize;
extern int vertices[MAX_VERTICES];
extern int mode;
extern int Starting_Locs[MAX_VERTICES];
extern double Current_Temp;

double Alpha;
double Start_Temp;
double End_Temp;
double Epoch_Length;
double Temp_Range;
double Exp_Tab[1000];
double Init_Prob,		/* acceptance rate for beginning temp        */
       Temp_Factor,		/* multiply Temp by this to decrease         */
       Size_Factor,		/* this many times num vertex moves per temp */
       Min_Percent,		/* <this % accepted ==> frozen		     */
       Epsilon;			/* temp will reduce to within this * End_Temp*/

static double Last_Time;
static int Hours;

void setup_exp_tab()
{
	int i;

	for (i=0; i<1000; i++) 
		Exp_Tab[i] = exp( -1.0 * ((i+1)/200.0)  );/**check!**/
}

int accept(double delta, double temp)
{
	int i = (int) ((200 * delta / temp) - 1);

	if ((i < 0) || ( (i<1000) && (rrandom() < Exp_Tab[i]) ))
	/*if ((i < 0) || ( (i<1000) && (rrandom() < exp (-1.0 * delta/temp))))*/
		return TRUE;
	return FALSE;
}

double find_delta(int v)
{
	extern LISTPTR nodes;
	extern int num_vertices[2];
	double curquot, nextquot, cur_val, next_val;
	int    cur_bal, next_bal;

	if (mode == MODE_MAX) {
		return ( (-1.0) *  (double) (nodes[v].cost));
	} else if (mode == MIN_QUOTIENT) {
		curquot  = (double) currentcost / currentminsize;
		nextquot = (double) (currentcost+nodes[v].cost) / 
				MIN(num_vertices[nodes[v].loc]-1,
				    num_vertices[nodes[v].loc]+1);
		return ( QUOT_MULT * (nextquot-curquot) );
	} else {
		cur_bal  = num - (2 * currentminsize);
		next_bal = num -  (2 * MIN(num_vertices[nodes[v].loc]-1,
				           num_vertices[nodes[v].loc]+1));
		cur_val  = currentcost + (Alpha * (cur_bal)); /* * cur_bal */
		next_val = (currentcost+nodes[v].cost) + 
				(Alpha *  (next_bal)); /* * next_bal */
		return ( next_val-cur_val );
	}
}

void shuffle_array(int a[])
{
        int i, j, temp;

        for (i=0; i<num; i++) {
                j = i + (random() % (num-i));
        	temp = a[i];
        	a[i] = a[j];
        	a[j] = temp;
	}
}

void starting_partition()
{
	extern LISTPTR nodes;
	extern int num_vertices[2];
	extern int node_cost[2][MAX_VERTICES];
	int i;

	for (i=0; i<num; i++)
		nodes[i].loc = Starting_Locs[i];
        currentcost = compute_costs();
        currentminsize = MIN(num_vertices[SET_A], num_vertices[SET_B]);
        reset_buckets(node_cost);
}

void one_temp(double temp, int *accepts, int *total, int exp_method)
{
	extern LISTPTR nodes;
	LISTPTR node;
	int i, index;
	double delta, r, exp_val;
	int e_index, move;  

	for (i=0; i<Epoch_Length; i++) {
		index = i % num;
		if (index == 0)
			shuffle_array(vertices);
		delta = find_delta(vertices[index]);
		e_index = (int) (200 * delta / temp) - 1;
		r=rrandom();
		if (exp_method == COMPUTE)
			exp_val = exp(-1*delta/temp); 
		else
			exp_val = ((e_index >= 0) && (e_index<1000)) 
				   	? Exp_Tab[e_index] : 0.0;
		if ((move = (r < exp_val)))
			(*accepts)++;
		(*total)++;
		if ((delta <= 0.0) || move) {
			node = &nodes[vertices[index]];
			delete_from_bucket(node);
		        bucket_move_vertex(node); 
               		insert_to_bucket(node->loc, -node->cost, node);
		}
	}
}

double find_start_temp(double goal_prob)
/****************************************************************************/
/** find_start_temp()	: 	  					   **/
/**			  This routine moves vertices between bucket	   **/
/**			  lists **without** updating Current_Max_Bucket[]  **/
/**			  or Current_Min_Bucket[].  The state of the buckets*/
/**			  after this routine isn't usable - reset_buckets()**/
/**			  with the held locs must be called (and is later) **/
/****************************************************************************/
{
	int    i, accepts, total;
	double search_ulimit;
	double search_llimit = 0.0;
	double temp;
	double lower_limit;
	double upper_limit;
	double accept_rate;
	double last_accept_rate = 0.0;

	search_ulimit = Start_Temp;
	temp = Start_Temp/ 2.0;
	accept_rate = 1.0;

	lower_limit = goal_prob - (.05 * goal_prob);
	upper_limit = goal_prob + (.05 * goal_prob);

	for (i=0; i<num; i++)
		vertices[i] = i;

	while ((accept_rate < lower_limit) || (accept_rate > upper_limit)) {
		accepts = total = 0;
		starting_partition();
		one_temp(temp, &accepts, &total, COMPUTE);
		accept_rate = accepts / (double) total;
		if (accept_rate < goal_prob)
			search_llimit = temp;
		else
			search_ulimit = temp;
		temp = (search_ulimit + search_llimit) / 2;
		if (temp < MIN_TEMP)
			return(MIN_TEMP);
		else if (search_ulimit <= search_llimit)
			return(temp);
		else if ((accept_rate < (goal_prob + upper_limit)) && 
			 ABS(accept_rate - last_accept_rate) < 
			 		   (Alpha * accept_rate))
			return(temp);
		last_accept_rate = accept_rate;
	}
	return temp;
}

int sort_and_tell(double *a, int limit, int start)
{
	int i,j, min;
	double data[5], temp;
	
	for (i=0; i<limit; i++) 
		data[i] = a[(start+i)%5];
	for (i=0; i<limit; i++) {
		min = i;	
		for (j=i; j<limit; j++) {
			if (data[j] < data[min])
				min = j;
		}
		if (i != min) {
			temp = data[i];
			data[i] = data[min];
			data[min] = temp;
		}
	}
	if (  (((mode == MODE_MAX) && (data[4] == tbestcost)) ||
	       ((mode != MODE_MAX) && (data[0] == tbestcost)))    &&
	      (data[0] < data[4]) )
		return TRUE;
	return FALSE;
}
			
int check_improvement()
{
	static int sol_index;
	static double solutions[5];

	sol_index = (sol_index + 1) % 5;
	solutions[sol_index] = currentcost;
	return sort_and_tell(solutions, 5, sol_index);
}

static int More_Runs;

void init_partitioning()
{
	extern int Geometric_Graphs, Line_Starts;
	extern int W_Starts;
	double unused;

	if (W_Starts) {
		init_walk(RESET_LOCS);
		one_walk(0,num, &unused);
	} else if (Geometric_Graphs && Line_Starts)
                line_randomize();
        else
                randomize_vertices();
	starting_partition();
	Current_Temp = Start_Temp;
}


double init_sim_annealing()
{
	extern int Geometric_Graphs, Line_Starts;
	extern int Line_Starts;
	extern int W_Starts;
	extern long seed;

	Hours = 0;
	Last_Time = 0.0;
	More_Runs = TRUE;
	randinit(seed);
	init_partitioning();
	if (Geometric_Graphs && (W_Starts || Line_Starts)) {
		Start_Temp = 5.0;
		Init_Prob = 0.02;
		Min_Percent = .001;
		Epsilon = .01;
	} else {
		Start_Temp = 10.0;
		Init_Prob = 0.4;
		Min_Percent = .02;
		Epsilon = .01;
	}
	Alpha = 0.05;
	Temp_Factor = 0.85;	/* will be reset after first annealing loop */
	Size_Factor = 16;
	Epoch_Length = Size_Factor * num;
	Start_Temp = find_start_temp(Init_Prob);
	starting_partition();
	return Start_Temp;
}

void sa_balance_partitioning()
{
	extern FILE *tfile;
	extern int hold2_num_vertices[2];
	extern int hold_kl_loc[MAX_VERTICES];

	init_local_opt(hold_kl_loc, hold2_num_vertices);
	balance_partitioning();
	if (!check_cost(NULL,  currentcost)) {
              fprintf(tfile,"*************************  annealing error #2 \n");
              printf("*************************  annealing error #2 \n");
	}
}

int termination_test(int first_done, double accept_rate, int iter)
{
	/*extern FILE *tfile;*/
	extern int Heuristic_Only;
	/*extern int tbestcost, tbestminsize;*/
	extern int num_passes;
	extern double tottime, cpu_seconds;
	int more_work = FALSE;

	if ((accept_rate >= Min_Percent) || check_improvement())
		more_work = TRUE;
	if (!first_done)
		return more_work;

	if (!((Heuristic_Only && (tottime <= cpu_seconds)) ||
	      (!Heuristic_Only && (iter <= num_passes)))) {
		more_work = FALSE;
		More_Runs = FALSE;
	}

	if (tottime >= (Last_Time+3600.0)) {
		Hours++;
		Last_Time = tottime;
		/*fprintf(tfile,"hours: %d, quot: %f\n", Hours,
			(double) tbestcost/tbestminsize);
		fflush(tfile);*/
	}
	return more_work;
}

int annealing_loop(int iter, int start_temp_factor)
/****************************************************************************/
/** annealing_loop()	: run from Start_Temp down to End_Temp (or no more **/
/**			  work).  This routine moves vertices between bucket*/
/**			  lists **without** updating Current_Max_Bucket[]  **/
/**			  or Current_Min_Bucket[].  The state of the buckets*/
/**			  after this routine isn't usable - reset_buckets()**/
/**			  with the held locs must be called (and is later) **/
/****************************************************************************/
{
	extern FILE *outfile;
	extern int Geometric_Graphs;
	extern int Heuristic_Only;
	extern int output_short, num_passes;
	extern double cpu_seconds, tottime;
	int accepts, total;
	int dummy;
	double temp_frac, num_temps;
	double accept_rate = 1.0, nonzero_ar = 1.0;
	double ave_time;
	int more_work = TRUE;

	log_time();
	while (more_work) { 
		iter++;
		accepts = total = 0;
		one_temp(Current_Temp, &accepts, &total, LOOKUP);
		accept_rate = accepts / (double) total;
		nonzero_ar = (accept_rate > 0.0) ? accept_rate : nonzero_ar;
		save_and_postproc(mode, 0, 0, &dummy, &dummy);
		more_work=termination_test(start_temp_factor,accept_rate,iter);
		Current_Temp = Current_Temp * Temp_Factor;
		time_ = log_time();
		tottime += time_;
		if(!output_short)
		   	print_iteration(iter+1,0,currentcost,currentminsize, 0);
	}
	if (!start_temp_factor) {
	   if ((mode == MODE_MAX) && (!Geometric_Graphs)) {
		ave_time = tottime / iter;
		temp_frac = Current_Temp / Start_Temp;
		if (Heuristic_Only) 
			num_temps=cpu_seconds/ave_time;
		else
			num_temps = num_passes;
		Temp_Factor = pow(temp_frac, 1.0/num_temps);
		Current_Temp = Start_Temp;
		Min_Percent = nonzero_ar;
		tottime = 0.0;
	   } else
		Temp_Factor = .95;
	   fprintf(outfile,"------------------------------------------"
	                   "--------------------------------------\n");
	}
	return iter;
}

void sim_annealing()
{
	extern FILE *tfile;
	/*extern int tbestcost;*/
	extern int Heuristic_Only;
	extern int Output_Sng_Distfile;
	extern int Num_Runs;
	extern int Done_With_Alg;
	/*int backup_start_cost = tbestcost;*/
	int flag = FALSE, iter=0;

	if (mode == MIN_QUOTIENT)
		mode = MODE_MIN;

	setup_exp_tab();
	Current_Temp = init_sim_annealing();

	Num_Runs = 0;
	annealing_loop(iter, flag);		/* reset Temp_Factor */
	/*tbestcost = backup_start_cost;*/
	flag=TRUE;
	while (More_Runs) {
		init_partitioning();
		Num_Runs++;
		if (Heuristic_Only)
			annealing_loop(0, flag);
		else
			iter = annealing_loop(iter, flag);
	}
	if (!check_cost(NULL,  currentcost)) {
              fprintf(tfile,"*************************  annealing error #1 \n");
              printf("*************************  annealing error #1 \n");
	}
	Done_With_Alg = TRUE; 
	if (mode == MODE_MIN) {
		mode = MIN_QUOTIENT;
		sa_balance_partitioning();
		/* untested */
        	if (Output_Sng_Distfile) 	/** set anything else up??? **/
        		 post_processing(TRUE, 0, 0);
	}
}
